# (c) Copyright Riverlane 2020-2025.
"""Datastructures for decoding syndromes."""

from __future__ import annotations

from collections import Counter, UserDict, defaultdict
from collections.abc import (
    Collection,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
)
from collections.abc import (
    Set as AbstractSet,
)
from functools import cached_property
from itertools import chain, repeat
from typing import Any, Literal, SupportsIndex, cast, overload

import numpy as np
from typing_extensions import Self

Bit = Literal[0, 1]


class DetectorRecord(UserDict):
    """Dictionary for recording information about a detector.
    String attributes for arbitrary values.
    Coordinate and time given as special data that is always
    defined.

    Parameters
    ----------
    spatial_coord : Tuple[int, ...], optional
        Variadic spatial coordinate of this detector, by default ().
    time : int, optional
        Time coordinate of this detector, by default 0.
    """

    def __init__(
        self, spatial_coord: tuple[float | int, ...] = (), time: int = 0, **kwargs
    ) -> None:
        super().__init__(spatial_coord=spatial_coord, time=time, **kwargs)

    @property
    def spatial_coord(self) -> tuple[float, ...]:
        """Spatial coordinate of this detector."""
        return self.data["spatial_coord"]

    @spatial_coord.setter
    def spatial_coord(self, value: tuple[float, ...]):
        self.data["spatial_coord"] = value

    @property
    def time(self) -> int:
        """Time coordinate of this detector."""
        return self.data["time"]

    @time.setter
    def time(self, value: int):
        self.data["time"] = value

    @property
    def full_coord(self) -> tuple[float, ...]:
        """Return the full coordinate for this detector, which is given in the form
        spatial coordinates then time coordinate.
        """
        return (*self.spatial_coord, self.time)

    @classmethod
    def from_dict(cls, property_dict: dict[str, Any]) -> DetectorRecord:
        """Create a DetectorRecord from a given property dict of optional values.
        This is included for compatibility for network-x vertex representations.

        Parameters
        ----------
        property_dict : Dict[str, Any]
            Any optional properties such as spatial_coord or time.

        Returns
        -------
        DetectorRecord
        """
        spatial_coord = property_dict.get("spatial_coord", ())
        # ensure a tuple
        spatial_coord = tuple(spatial_coord)
        time = property_dict.get("time", 0)
        return DetectorRecord(time=time, spatial_coord=spatial_coord)

    @classmethod
    def from_sequence(cls, sequence: Sequence[float | int]) -> DetectorRecord:
        """Create a DetectorRecord given a sequence if floats or ints. The
        first n-1 items in the sequence are used for the spatial coordinate
        and the nth item is used as the time coordinate.

        Parameters
        ----------
        sequence : Sequence[float  |  int]
            A sequence of floats or ints specifying the coordinates.

        Returns
        -------
        DetectorRecord
        """
        if len(sequence) == 0:
            return DetectorRecord()
        *spatial_coord, time = sequence
        return DetectorRecord(spatial_coord=tuple(spatial_coord), time=int(time))


class OrderedSyndrome(Sequence[int], AbstractSet[int]):
    """Immutable ordered mod 2 set of detectors. Where detectors
    are represented by integers.

    Parameters
    ----------
    detectors : Iterable[int], optional
        Detectors to put into this ordered mod 2 collection, by default ().
    enforce_mod_2 : bool, optional
        Whether to enforce that there's only an odd number of each detector in
        the syndrome, by default True.
    """

    def __init__(self, detectors: Iterable[int] = (), enforce_mod_2: bool = True):
        if enforce_mod_2:
            detector_counts = Counter(detectors)
            self._detectors = dict.fromkeys(
                [edge for edge, count in detector_counts.items() if count % 2 == 1]
            )
        else:
            self._detectors = dict.fromkeys(detectors)

    @cached_property
    def _as_tuple(self) -> tuple[int, ...]:
        """Defined to create immutable object to hash, and to make `__getitem__` a O(1)
        method. Exists as a property to avoid duplication of data in core member
        attributes, and to have this be created only when needed.
        """
        return tuple(self._detectors)

    def __len__(self) -> int:
        return len(self._detectors)

    def __getitem__(self, index):
        return self._as_tuple.__getitem__(index)

    def __contains__(self, x: object) -> bool:
        return self._detectors.__contains__(x)

    def __iter__(self) -> Iterator[int]:
        return self._detectors.__iter__()

    def __repr__(self) -> str:
        return f"[{', '.join(map(str, self))}]"

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: object) -> bool:
        # Could allow other sequences to be comparable?
        return isinstance(other, OrderedSyndrome) and self._as_tuple == other._as_tuple

    @classmethod
    def from_bitstring(cls, bitstring: Sequence[Bit]) -> OrderedSyndrome:
        """Create an ordered syndrome from a bistring.

        Parameters
        ----------
        bitstring : Sequence[Bit]
            Bitstring to convert from.

        Returns
        -------
        OrderedSyndrome
            Indices of the non zero elements in the bitstring as a syndrome.
        """
        return OrderedSyndrome(np.flatnonzero(bitstring).tolist())

    def as_bitstring(self, num_bits: int) -> list[Bit]:
        """Convert OrderedSyndrome to a bitstring, with `num_bits` bits."""
        syndrome_bits = np.zeros(num_bits, dtype=np.uint8)
        syndrome_bits[list(self._detectors)] = 1
        result = syndrome_bits.tolist()
        return cast(list[Bit], result)

    def split_at_symptom(
        self, split_detection_event: int
    ) -> tuple[list[int], list[int]]:
        """Split OrderedSyndrome in to two lists of detection events in
        order, ending the first at the given detector event.

        Parameters
        ----------
        split_detection_event : int
            The largest detection event symptom.
            Events after this are interpreted to be leakage heralding events.

        Returns
        -------
        Tuple[List[int], List[int]]
            Lists of post and pre symptom detection events.
        """
        pre_dets, post_dets = [], []
        for event_id in self:
            if event_id > split_detection_event:
                post_dets.append(event_id - split_detection_event - 1)
            else:
                pre_dets.append(event_id)
        return pre_dets, post_dets

    def split_by_time_coord(
        self, detector_records: Mapping[int, DetectorRecord], layers: int
    ) -> list[OrderedSyndrome]:
        """Split the OrderedSyndrome into a time-ordered list of OrderedSyndromes, with
        `layers` number of items in the list.
        """
        syndromes_by_time: dict[int, list[int]] = defaultdict(list)
        for detector in self:
            detector_time = detector_records[detector].time
            syndromes_by_time[detector_time].append(detector)

        result = [OrderedSyndrome() for _ in range(layers)]
        for time, syndromes in syndromes_by_time.items():
            result[time] = OrderedSyndrome(syndromes)

        return result

    def as_layers(
        self, syndromes_per_layer: int | list[int], total_layers: int | None = None
    ) -> list[list[int]]:
        """Create a sequence of layers from a syndrome, where each layer is a
        collection of integers representing the detectors triggered on that layer.
        Each layer should be contiguously indexed from 0.

        Parameters
        ----------
        syndromes_per_layer : Union[int, List[int]]
            Number of possible detectors on each layer. If int, assume all layers
            are the same size.
        total_layers : Optional[int]
            Number of layers returned. If None, use len(syndromes_per_layer).
            If syndromes_per_layer is int, use however many layers is necessary to fit
            all the syndromes. By default, None.
        """
        if total_layers is None:
            if isinstance(syndromes_per_layer, int):
                total_layers = (
                    max(self._detectors, default=-1) // syndromes_per_layer
                ) + 1
            else:
                total_layers = len(syndromes_per_layer)
        if isinstance(syndromes_per_layer, int):
            syndromes_per_layer = [syndromes_per_layer for _ in range(total_layers)]
        layers: list[list[int]] = [[] for _ in range(total_layers)]

        if len(syndromes_per_layer) == 0:
            return layers
        current_layer = 0
        syndromes_before_current_layer = 0
        syndromes_including_current_layer = syndromes_per_layer[0]
        for symptom in sorted(self._detectors):
            # check if crossed to new layer
            while syndromes_including_current_layer <= symptom:
                syndromes_before_current_layer += syndromes_per_layer[current_layer]
                current_layer += 1
                syndromes_including_current_layer += syndromes_per_layer[current_layer]
            layers[current_layer].append(symptom - syndromes_before_current_layer)

        return layers

    @classmethod
    def from_layers(
        cls, layers: Sequence[Iterable[int]], syndromes_per_layer: int | list[int]
    ) -> OrderedSyndrome:
        """Create a syndrome from a sequence of layers, where each layer is a
        collection of integers representing the detectors triggered on that layer.
        Each layer should be contiguously indexed from 0.

        Parameters
        ----------
        layers : Sequence[Iterable[int]]
            Sequence of layers of positive detectors.
        syndromes_per_layer : Union[int, List[int]]
            Number of possible detectors on each layer. If int, assume all layers same
            size.

        Returns
        -------
        OrderedSyndrome
        """
        syndrome: list[int] = []
        layer_synd_add = 0
        for i, layer in enumerate(layers):
            syndrome.extend([synd + layer_synd_add for synd in layer])
            if isinstance(syndromes_per_layer, int):
                layer_synd_add += syndromes_per_layer
            else:
                layer_synd_add += syndromes_per_layer[i]
        return cls(syndrome)


class Bitstring:
    """Class which efficiently represents a bitstring."""

    __slots__ = ("_bits",)

    def __init__(self, value: int = 0):
        if value < 0:
            msg = "Bitstring cannot be a negative value."
            raise ValueError(msg)
        self._bits = value

    @classmethod
    def from_bits(cls, bits: Iterable[Bit]):
        """Create a bitstring from a list of bits. The bits are
        assumed to be in little endian form, i.e. the lsb is the first item
        in the iterable.

        >>> Bitstring.from_bits([1, 0, 0, 1, 1])
        0b11001
        """
        value = 0
        for bit_num, bit in enumerate(bits):
            value |= bit << bit_num
        return cls(value)

    @classmethod
    def from_bytes(cls, bytes_: Collection[SupportsIndex]):
        """Create a bitstring from a collection of bytes. For example, a numpy array of
        uint8s. Assumed to be in little endian form.

        >>> Bitstring.from_bytes(np.array([2, 128], dtype=np.uint8))
        0b1000000000000010
        """
        return cls(int.from_bytes(bytes_, byteorder="little"))

    @classmethod
    def from_indices(cls, indices: Iterable[int]):
        """Create a bitstring from a list of non-zero indices.

        >>> Bitstring.from_indices([0, 2, 5])
        0b100101
        """
        value = 0
        for index in indices:
            value |= 1 << index
        return cls(value)

    def to_indices(self) -> list[int]:
        """Create a list of indices of the non-zero elements in this bitstring

        Returns
        -------
        List[int]
            The indices of non-zero elements in the bitstring.
        """
        return [index for index, value in enumerate(self) if value == 1]

    def to_words(self, num_bits_per_word: int) -> Iterator[FixedWidthBitstring]:
        """Convert the bitstring into words of a given length."""
        for shift in range(0, len(self), num_bits_per_word):
            yield FixedWidthBitstring(num_bits_per_word, self._bits >> shift)

    def bit_count(self) -> int:
        """Get the number of 1s in this bitstring."""
        return self._bits.bit_count()

    def __or__(self, rhs: object) -> Bitstring:
        if isinstance(rhs, Bitstring):
            return Bitstring(self._bits | rhs._bits)
        return NotImplemented

    def __ior__(self, rhs: object) -> Self:
        if isinstance(rhs, Bitstring):
            self._bits |= rhs._bits
            return self
        return NotImplemented

    def __and__(self, rhs: object) -> Bitstring:
        if isinstance(rhs, Bitstring):
            return Bitstring(self._bits & rhs._bits)
        return NotImplemented

    def __iand__(self, rhs: object) -> Self:
        if isinstance(rhs, Bitstring):
            self._bits &= rhs._bits
            return self
        return NotImplemented

    def __xor__(self, rhs: object) -> Bitstring:
        if isinstance(rhs, Bitstring):
            return Bitstring(self._bits ^ rhs._bits)
        return NotImplemented

    def __ixor__(self, rhs: object) -> Self:
        if isinstance(rhs, Bitstring):
            self._bits ^= rhs._bits
            return self
        return NotImplemented

    def __lshift__(self, rhs: object) -> Bitstring:
        if isinstance(rhs, int):
            return Bitstring(self._bits.__lshift__(rhs))
        return NotImplemented

    def __rshift__(self, rhs: object) -> Bitstring:
        if isinstance(rhs, int):
            return Bitstring(self._bits.__rshift__(rhs))
        return NotImplemented

    def __iter__(self) -> Iterator[Bit]:
        # max is required so that list(Bitstring(0)) == [0] and not []
        for i in range(max(1, self._bits.bit_length())):
            yield (self._bits >> i) & 1  # type: ignore[misc]

    def __reversed__(self) -> Iterator[Bit]:
        # max is required so that list(Bitstring(0)) == [0] and not []
        for i in reversed(range(max(1, self._bits.bit_length()))):
            yield (self._bits >> i) & 1  # type: ignore[misc]

    @overload
    def __getitem__(self, index: int) -> Bit: ...

    @overload
    def __getitem__(self, index: slice) -> Bitstring: ...

    def __getitem__(self, index: int | slice) -> Bit | Bitstring:
        if isinstance(index, int):
            return cast(Literal[0, 1], (self._bits >> index) & 1)
        if isinstance(index, slice):
            start, stop, _ = index.indices(len(self))
            mask = (1 << (stop - start)) - 1
            return Bitstring((self._bits >> start) & mask)
        msg = f"Bitstring indices must be integers or slices, not {type(index)}"
        raise TypeError(msg)

    def __eq__(self, rhs: object) -> bool:
        if isinstance(rhs, Bitstring):
            return self._bits == rhs._bits
        return NotImplemented

    def __len__(self) -> int:
        return max(1, self._bits.bit_length())

    def __index__(self):
        return self._bits

    def __repr__(self) -> str:
        return bin(self._bits)

    def __hash__(self) -> int:
        return hash(self._bits)


class FixedWidthBitstring(Bitstring):
    """Class which represents a fixed precision bitstring."""

    __slots__ = ("_width",)

    def __init__(self, width: int, value: int = 0):
        if width < 1:
            msg = "Width of bitstring must be greater than zero."
            raise ValueError(msg)
        super().__init__(value & ((1 << width) - 1))
        self._width = width

    @classmethod
    def from_indices(cls, indices: Iterable[int]):
        value = 0
        for index in indices:
            value |= 1 << index
        return cls(value.bit_length(), value)

    @classmethod
    def from_bits(cls, bits: Iterable[Bit]):
        value = 0
        # Use variable shadowing in case the enumerate block is never entered
        bit_num = -1
        for bit_num, bit in enumerate(bits):
            value |= bit << bit_num
        return cls(bit_num + 1, value)

    @classmethod
    def from_bytes(cls, bytes_: Collection[SupportsIndex]):
        return cls(len(bytes_) * 8, int.from_bytes(bytes_, byteorder="little"))

    def change_width(self, new_width: int):
        """Change the width of this bitstring internally. If the new width is
        less than the current width then any excess bits will be removed.
        """
        if new_width < self._width:
            self._bits &= (1 << new_width) - 1
        self._width = new_width

    def __or__(self, rhs: object) -> FixedWidthBitstring:
        if isinstance(rhs, Bitstring):
            return FixedWidthBitstring(self._width, self._bits | rhs._bits)
        return NotImplemented

    def __ior__(self, rhs: object) -> Self:
        if isinstance(rhs, Bitstring):
            # Need the mask here to remove the upper bits if rhs has more
            # bits than self.
            self._bits |= rhs._bits & ((1 << self._width) - 1)
            return self
        return NotImplemented

    def __xor__(self, rhs: object) -> FixedWidthBitstring:
        if isinstance(rhs, Bitstring):
            return FixedWidthBitstring(self._width, self._bits ^ rhs._bits)
        return NotImplemented

    def __ixor__(self, rhs: object) -> Self:
        if isinstance(rhs, Bitstring):
            # Need the mask here to remove the upper bits if rhs has more
            # bits than self.
            self._bits ^= rhs._bits & ((1 << self._width) - 1)
            return self
        return NotImplemented

    def __and__(self, rhs: object) -> FixedWidthBitstring:
        if isinstance(rhs, Bitstring):
            return FixedWidthBitstring(self._width, self._bits & rhs._bits)
        return NotImplemented

    def __iand__(self, rhs: object) -> Self:
        if isinstance(rhs, Bitstring):
            # Need the mask here to remove the upper bits if rhs has more
            # bits than self.
            self._bits &= rhs._bits & ((1 << self._width) - 1)
            return self
        return NotImplemented

    def __invert__(self) -> FixedWidthBitstring:
        return FixedWidthBitstring(self._width, ~self._bits)

    def __iter__(self) -> Iterator[Bit]:
        yield from chain(
            super().__iter__(),
            cast(
                Iterator[Literal[0]],
                repeat(0, self._width - max(1, self._bits.bit_length())),
            ),
        )

    def __reversed__(self) -> Iterator[Bit]:
        yield from chain(
            cast(
                Iterator[Literal[0]],
                repeat(0, self._width - max(1, self._bits.bit_length())),
            ),
            super().__reversed__(),
        )

    def __len__(self) -> int:
        return self._width

    @overload
    def __getitem__(self, index: int) -> Bit: ...

    @overload
    def __getitem__(self, index: slice) -> Bitstring: ...

    def __getitem__(self, index: int | slice) -> Bit | Bitstring:
        if isinstance(index, int):
            return cast(Literal[0, 1], (self._bits >> index) & 1)
        if isinstance(index, slice):
            start, stop, _ = index.indices(len(self))
            mask = (1 << (stop - start)) - 1
            return FixedWidthBitstring(
                width=stop - start, value=(self._bits >> start) & mask
            )
        msg = f"Bitstring indices must be integers or slices, not {type(index)}"
        raise TypeError(msg)

    def __add__(self, rhs: object) -> FixedWidthBitstring:
        if isinstance(rhs, FixedWidthBitstring):
            return FixedWidthBitstring(
                self._width + rhs._width,
                (self._bits << rhs._width) | rhs._bits,
            )
        return NotImplemented


def get_round_words(
    detector_indices: Sequence[int], num_detectors_per_round: int, num_rounds: int
) -> Iterator[FixedWidthBitstring]:
    """For given sparse-form sequence, return the corresponding bit-strings
    that correspond to individual decoding rounds.

    The sparse sequence is expected to contain defects, boundary adjacency, logical
    adjacency or other bits that are specified per-detector.

    Parameters
    ----------
    detector_indices : Sequence[int]
        Sparse form of detector aligned data.
    num_detectors_per_round : int
        Number of detectors in each decoding round.
    num_rounds : int
        Number of decoding rounds.

    Yields
    ------
    Iterator[FixedWidthBitstring]
        Bitstrings for each decoding round, containing dense-form data.
    """

    total_width = num_detectors_per_round * num_rounds
    if len(detector_indices) == 0:
        bitstring = FixedWidthBitstring(total_width)
    else:
        bitstring = FixedWidthBitstring.from_indices(detector_indices)
        bitstring.change_width(total_width)

    return bitstring.to_words(num_detectors_per_round)



# (c) Copyright Riverlane 2020-2025.
import csv
from collections.abc import Iterable, Iterator
from itertools import cycle
from pathlib import Path

from deltakit_core.decoding_graphs import Bitstring


def split_input_data_to_c64(
    input_data: Iterable[Iterable[Bitstring]], output_path: Path
):
    """
    Given a list of round-split decoder input data, this will output them to a file
    as comma separated 64-bit integers. Each line is a new shot.

    The decoder input data could be measurements or syndromes.
    """
    with output_path.open("w", newline="", encoding="ascii") as output_handle:
        cw = csv.writer(output_handle)
        for input_datum in input_data:
            words = [
                int(item)
                for line in [
                    input_datum_round.to_words(64) for input_datum_round in input_datum
                ]
                for item in line
            ]
            cw.writerow(words)


def c64_to_addressed_input_words(
    input_path: Path, round_width: int
) -> Iterator[list[tuple[int, int]]]:
    """Given a path to a c64 file, return, for each line, the words within the line
    and the index of each word in each round. These are appropriate for input to a
    hardware decoder, they could be syndromes or measurement data.

    Parameters
    ----------
    input_path : Path
        Path to the c64 file to read.
    round_width : int
        The number of detectors or measurements per round.

    Yields
    ------
    List[Tuple[int, int]]
        List of addressed 64bit words
    """
    with input_path.open("r", encoding="ascii") as input_handle:
        syndrome_reader = csv.reader(input_handle)
        for row in syndrome_reader:
            round_indices = cycle(range(round_width // 64 + 1))
            yield [
                (word_index, int(word_64))
                for word_index, word_64 in zip(round_indices, row, strict=False)
            ]



# (c) Copyright Riverlane 2020-2025.
"""Module for parsing detector error models."""

from __future__ import annotations

import collections.abc
import warnings
from collections import Counter
from collections.abc import Iterable, Iterator
from itertools import chain, zip_longest
from typing import Generic, Protocol, TypeVar, cast

try:
    import lestim as stim
except ImportError:
    import stim


from deltakit_core.decoding_graphs._data_qubits import (
    DecodingEdge,
    DecodingHyperEdge,
    EdgeRecord,
)
from deltakit_core.decoding_graphs._decoding_graph import (
    _QECNX,
    DecodingHyperGraph,
    NXDecodingGraph,
)
from deltakit_core.decoding_graphs._syndromes import DetectorRecord


class CoordinateOffset(tuple):  # noqa: PLW1641
    """Class to track the coordinate offset in a Detector Error Model."""

    def __new__(cls, offset: Iterable[float | int] = ()):
        return super().__new__(cls, offset)

    def __add__(self, other: object) -> CoordinateOffset:
        if isinstance(other, collections.abc.Iterable):
            return CoordinateOffset(
                self_offset + other_offset
                for self_offset, other_offset in zip_longest(self, other, fillvalue=0)
            )
        return NotImplemented

    def __radd__(self, other: object) -> CoordinateOffset:
        return self.__add__(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CoordinateOffset):
            return super().__eq__(other)
        return NotImplemented


class ErrorHandler(Protocol):
    """General callable protocol used in the DEM parser to handler stim's
    error instructions."""

    def __call__(self, error: stim.DemInstruction, detector_offset: int) -> None: ...


class DetectorHandler(Protocol):
    """General callable protocol used in the DEM parser to handle stim's
    detector instructions."""

    def __call__(
        self,
        detector: stim.DemInstruction,
        detector_offset: int,
        coordinate_offset: CoordinateOffset,
    ) -> None: ...


class ObservableHandler(Protocol):
    """General callable protocol used in the DEM parser to handler stim's
    logical_observable instructions."""

    def __call__(self, logical_observable: stim.DemInstruction) -> None: ...


class DetectorRecorder(DetectorHandler):
    """A type of detector handler which populates a dictionary of
    DetectorRecords."""

    def __init__(self) -> None:
        self._detector_records: dict[int, DetectorRecord] = {}

    @property
    def detector_records(self) -> dict[int, DetectorRecord]:
        """Get the detector records for this class."""
        return self._detector_records

    def __call__(
        self,
        detector: stim.DemInstruction,
        detector_offset: int,
        coordinate_offset: CoordinateOffset,
    ):
        detector_index = (
            detector_offset + cast(list[stim.DemTarget], detector.targets_copy())[0].val
        )
        detector_coordinate = coordinate_offset + detector.args_copy()
        self._detector_records[detector_index] = DetectorRecord.from_sequence(
            detector_coordinate
        )


def observable_warning(logical_observable: stim.DemInstruction) -> None:
    """Type of logical observable handler which raises a warning when this
    instruction is encountered."""
    logicals = " ".join(
        f"L{target.val}"
        for target in cast(list[stim.DemTarget], logical_observable.targets_copy())
    )
    warnings.warn(
        f"Isolated logical observables {logicals} declared in DEM file.",
        UserWarning,
        stacklevel=2,
    )


def collect_edges(
    error: stim.DemInstruction, detector_offset: int
) -> Iterator[tuple[set[int], float, list[int]]]:
    """Iterate through the DEM targets in the error and yield all the
    detectors which make up the edge, the probability of an error on this edge
    and the logical observables which would be affected by an error on the
    edge.

    Edges are yielded when the target is a target separator.

    Parameters
    ----------
    error : stim.DemInstruction
        A DemInstruction with type as "error"
    detector_offset : int
        The offset to add to all detectors in the edge

    Yields
    ------
    Iterator[Tuple[Set[int], float, int | None]]
        Set[int]:
            These are the detectors which are in the edge
        float:
            This is the probability of an error on this edge
        List[int]:
            List of which logicals are affected by errors on this edge. If no
            logical is affected this will be an empty list.
    """
    p_err = error.args_copy()[0]
    vertices_in_edge: set[int] = set()
    logicals_affected: list[int] = []
    for target in chain(
        cast(list[stim.DemTarget], error.targets_copy()), (stim.target_separator(),)
    ):
        if target.is_separator():
            yield vertices_in_edge, p_err, logicals_affected
            vertices_in_edge = set()
            logicals_affected = []
        elif target.is_relative_detector_id():
            vertices_in_edge.add(detector_offset + target.val)
        elif target.is_logical_observable_id():
            logicals_affected.append(target.val)


class LogicalsInEdges(ErrorHandler):
    """A type of edge handler which collects the errors into a set of edges
    and edges which affect the logicals."""

    def __init__(self, num_logicals: int):
        self._edges: set[DecodingHyperEdge] = set()
        self._edge_records: dict[DecodingHyperEdge, EdgeRecord] = {}
        self._logicals: list[set[DecodingHyperEdge]] = [
            set() for _ in range(num_logicals)
        ]

    @property
    def edges(self) -> set[DecodingHyperEdge]:
        """Gets all edges from a detector error model."""
        return self._edges

    @property
    def logicals(self) -> list[set[DecodingHyperEdge]]:
        """Get the edges which affect each logical of a detector error
        model."""
        return self._logicals

    def __call__(self, error: stim.DemInstruction, detector_offset: int):
        for vertices, p_err, logicals in collect_edges(error, detector_offset):
            edge = DecodingHyperEdge(vertices)
            self._edge_records[edge] = EdgeRecord(p_err=p_err)
            self._edges.add(edge)
            for logical in logicals:
                self._logicals[logical].add(edge)


EH = TypeVar("EH", bound=ErrorHandler)
DH = TypeVar("DH", bound=DetectorHandler)


class DemParser(Generic[EH, DH]):
    """Class which handles parsing a detector error model. This class tracks
    the detector and coordinate offsets as member data while parsing and passes
    this information into the relevant handlers.

    The error handler is a function which takes a "error" type DEM instruction
    and the current detector offset and tracks the state of the errors.

    The detector handler is a function which takes a "detector" type DEM
    instruction, the current detector offset and the coordinate offset and
    tracks the detectors.
    """

    def __init__(
        self,
        error_handler: EH,
        detector_handler: DH,
        observable_handler: ObservableHandler = observable_warning,
    ):
        self._error_handler = error_handler
        self._detector_handler = detector_handler
        self._observable_handler = observable_handler
        self._detector_offset = 0
        self._coordinate_offset = CoordinateOffset()

    @property
    def error_handler(self) -> EH:
        """Get the error handler for this DEM parser."""
        return self._error_handler

    @property
    def detector_handler(self) -> DH:
        """Get the detector handler for this DEM parser."""
        return self._detector_handler

    @property
    def observable_handler(self) -> ObservableHandler:
        """Get the logical observable handler for this DEM parser."""
        return self._observable_handler

    def parse(self, detector_error_model: stim.DetectorErrorModel):
        """Parse a detector error model using the given error and detector
        handlers. The class keeps track of the detector and coordinate offsets.

        Parameters
        ----------
        detector_error_model : stim.DetectorErrorModel
            A stim detector error model.
        """
        for instruction in detector_error_model:
            if isinstance(instruction, stim.DemRepeatBlock):
                for _ in range(instruction.repeat_count):
                    self.parse(instruction.body_copy())
            elif isinstance(instruction, stim.DemInstruction):
                if instruction.type == "error":
                    self._error_handler(instruction, self._detector_offset)
                elif instruction.type == "shift_detectors":
                    self._detector_offset += cast(int, instruction.targets_copy()[0])
                    self._coordinate_offset += instruction.args_copy()
                elif instruction.type == "detector":
                    self._detector_handler(
                        instruction, self._detector_offset, self._coordinate_offset
                    )
                elif instruction.type == "logical_observable":
                    self._observable_handler(instruction)


def dem_to_hypergraph_and_logicals(
    dem: stim.DetectorErrorModel,
) -> tuple[DecodingHyperGraph, list[set[DecodingHyperEdge]]]:
    """Convert a Stim detector error model into a DecodingHyperGraph and a
    list of edges which affect the logical observable at the index in the list.

    Parameters
    ----------
    dem : stim.DetectorErrorModel
        Stim detector error model to convert.

    Returns
    -------
    Tuple[DecodingHyperGraph, List[Set[DecodingHyperEdge]]]
        DecodingHyperGraph created from detector error model and a list of
        edges which affect the logical.
    """
    edge_records: dict[DecodingHyperEdge, EdgeRecord] = {}
    logicals_affected: list[set[DecodingHyperEdge]] = [
        set() for _ in range(dem.num_observables)
    ]

    def error_handler_no_multiedges(error: stim.DemInstruction, detector_offset: int):
        """Type of error handler for constructing decoding hypergraph and
        logicals hyperedges which are not multiedges from error instructions."""
        for vertices, p_err, logicals in collect_edges(error, detector_offset):
            hyperedge = DecodingHyperEdge(vertices)
            old_hyperedge_record = edge_records.get(hyperedge)
            if old_hyperedge_record is not None:
                p_err = p_err * (  # noqa: PLW2901
                    1 - old_hyperedge_record.p_err
                ) + old_hyperedge_record.p_err * (1 - p_err)
                old_hyperedge_record.p_err = p_err
            edge_records[hyperedge] = EdgeRecord(p_err=p_err)
            for logical_index in logicals:
                logicals_affected[logical_index].add(hyperedge)

    parser = DemParser(
        error_handler_no_multiedges, DetectorRecorder(), observable_warning
    )
    parser.parse(dem)

    return (
        DecodingHyperGraph(
            list(edge_records.items()),
            parser.detector_handler.detector_records,
        ),
        logicals_affected,
    )


def dem_to_decoding_graph_and_logicals(
    dem: stim.DetectorErrorModel,
) -> tuple[NXDecodingGraph, list[set[DecodingEdge]]]:
    """Convert a detector error model into a NXDecodingGraph object and a list
    of edges which effect particular logicals.

    Parameters
    ----------
    dem : stim.DetectorErrorModel
        The stim Detector Error Model to convert into a graph. The edges
        should already be decomposed into edges of max degree two.

    Returns
    -------
    Tuple[NXDecodingGraph, List[Set[DecodingEdge]]]
        A decoding graph and a list of edges which effect each logical.

    Raises
    ------
    ValueError
        If while parsing the DEM it encounters an edge of degree greater than
        two.
    """
    boundary = dem.num_detectors
    graph = _QECNX()
    graph.add_node(boundary, **DetectorRecord((-1, -1), 0))
    logicals_affected: list[set[DecodingEdge]] = [
        set() for _ in range(dem.num_observables)
    ]

    def error_handler(error: stim.DemInstruction, detector_offset: int):
        """Type of error handler for constructing decoding graph and
        logical edges from error instructions."""
        degree_target = 2
        for vertices, p_err, logicals in collect_edges(error, detector_offset):
            if (degree := len(vertices)) == 0:
                warnings.warn(
                    "Degree 0 edge has been skipped over in graph creation.",
                    UserWarning,
                    stacklevel=2,
                )
                continue
            if degree > degree_target:
                msg = f"Edge of degree {degree} cannot be converted to decoding edge."
                raise ValueError(msg)

            u, v = (next(iter(vertices)), boundary) if degree == 1 else tuple(vertices)
            old_edge_data = (
                edge_data
                if (edge_data := graph.get_edge_data(u, v)) is not None
                else None
            )
            if old_edge_data is not None:
                old_p_err = old_edge_data["p_err"]
                p_err = p_err * (1 - old_p_err) + old_p_err * (1 - p_err)  # noqa: PLW2901

            edge = DecodingEdge(u, v)
            edge_record = EdgeRecord(p_err=p_err)
            graph.add_edge(u, v, **edge_record)
            for logical_index in logicals:
                logicals_affected[logical_index].add(edge)

    def detector_handler(
        detector: stim.DemInstruction,
        detector_offset: int,
        coordinate_offset: CoordinateOffset,
    ):
        """Type of detector handler for adding coordinate annotated detectors
        to the decoding graph."""
        detector_index = (
            detector_offset + cast(list[stim.DemTarget], detector.targets_copy())[0].val
        )
        detector_coordinate = coordinate_offset + detector.args_copy()
        graph.add_node(
            detector_index, **DetectorRecord.from_sequence(detector_coordinate)
        )

    parser = DemParser(error_handler, detector_handler, observable_warning)
    parser.parse(dem)

    return NXDecodingGraph(graph, [boundary]), logicals_affected


class DetectorCounter(ErrorHandler):
    """Error handler that counts the number of detectors that each error flips."""

    def __init__(self) -> None:
        self.counter: Counter[int] = Counter()

    def __call__(self, error: stim.DemInstruction, detector_offset: int) -> None:
        self.counter.update(
            len(detectors) for detectors, _, _ in collect_edges(error, detector_offset)
        )

    def max_num_detectors(self) -> int:
        """
        Returns
        -------
        int
            Maximum number of detectors associated with a single error.
        """
        return max(self.counter.keys(), default=0)



# (c) Copyright Riverlane 2020-2025.
"""Module containing useful functions to create and interact with decoding graphs."""

import decimal
import json
import math
from collections.abc import Iterable
from collections.abc import Set as AbstractSet
from typing import no_type_check

import networkx as nx
import numpy as np

from deltakit_core.decoding_graphs._decoding_graph import (
    AnyEdgeT,
    DecodingEdge,
    DecodingHyperGraph,
    DecodingHyperMultiGraph,
    DetectorRecord,
    EdgeRecord,
    HyperLogicals,
    HyperMultiGraph,
    NXDecodingGraph,
    NXDecodingMultiGraph,
    NXLogicals,
)
from deltakit_core.decoding_graphs._dem_parsing import (
    DemParser,
    DetectorCounter,
    dem_to_decoding_graph_and_logicals,
)

try:
    import lestim as stim
except ImportError:
    import stim


def filter_to_data_edges(graph: NXDecodingGraph) -> list[DecodingEdge]:
    """Given some decoding edges, filter to those that directly correspond to a data
    qubit.
    """
    return [edge for edge in graph.edges if edge.is_spacelike(graph.detector_records)]


def filter_to_measure_edges(graph: NXDecodingGraph) -> list[DecodingEdge]:
    """Given some decoding edges, filter to those that directly correspond to a
    measurement event, i.e. they only move in the time axis.
    """
    return [edge for edge in graph.edges if edge.is_timelike(graph.detector_records)]


def hypergraph_to_weighted_edge_list(
    hypergraph: DecodingHyperGraph,
) -> list[tuple[frozenset[int], float]]:
    """Return a weighted edge list representation of a decoding hypergraph.

    A format that is useful for integration with other libraries for example in
    Belief Propagation decoders.

    Returns
    -------
    List[Tuple[frozenset[int], float]]]
        List of edges as a set of nodes and error probabilities.
        Ex. `[({0, 3, 5}, 0.05)]` represents a hyperedge with nodes
        {0, 3, 5} and error probability 0.05.
    """
    return [
        (edge.vertices, hypergraph.edge_records[edge].p_err)
        for edge in hypergraph.edges
    ]


def has_contiguous_nodes(graph: HyperMultiGraph, start: int = 0) -> bool:
    """Returns whether the nodes of the graph have contiguous indices with
    respect to the start index.

    Parameters
    ----------
    graph : HyperMultiGraph
        Any graph.
    start : int, optional
        Which index to start counting nodes from at, by default 0.

    Returns
    -------
    bool
        Whether all nodes in the graph are zero indexed and contiguous.
    """
    return all(index == node for index, node in enumerate(graph.nodes, start))


def single_boundary_is_last_node(graph: HyperMultiGraph) -> bool:
    """
    Parameters
    ----------
    graph : HyperMultiGraph
        Any graph

    Returns
    -------
    bool
        True if graph has a single boundary and that boundary is the last node where
        nodes are ordered wrt their index.
    """
    return (
        len(graph.boundaries) == 1 and next(iter(graph.boundaries)) == graph.nodes[-1]
    )


def is_single_connected_component(graph: NXDecodingGraph) -> bool:
    """
    Determines whether a graph consists of a single connected component
    excluding the boundary nodes of the graph.

    Parameters
    ----------
    graph : NXDecodingGraph
        A NetworkX decoding graph

    Returns
    -------
    bool
        True if the graph contains a single connected component False otherwise.
    """
    return nx.number_connected_components(graph.no_boundary_view) == 1


def parse_stim_circuit(
    stim_circuit: stim.Circuit,
) -> tuple[NXDecodingGraph, list[set[DecodingEdge]], stim.Circuit]:
    """Parse a Stim file into a decoding graph and the relevant logicals.

    Parameters
    ----------
    stim_circuit : stim.Circuit
        Input Stim circuit to parse.

    Returns
    -------
    Tuple[NXDecodingGraph, List[Set[DecodingEdge]], stim.Circuit]
        The decoding graph, the logicals, and the Stim circuit. The Stim
        circuit will be unchanged.
    """

    dem = stim_circuit_to_graph_dem(stim_circuit)
    graph, logicals = dem_to_decoding_graph_and_logicals(dem)
    return graph, logicals, stim_circuit


def stim_circuit_to_graph_dem(
    stim_circuit: stim.Circuit, approximate_disjoint_errors: bool = True
) -> stim.DetectorErrorModel:
    """For a given stim circuit, return the graph-like detector error model.
    If the non-decomposed DEM is graph-like, that will be returned. Otherwise,
    the decomposed DEM will be returned.

    Parameters
    ----------
    stim_circuit : stim.Circuit
        Stim circuit to get the DEM for
    approximate_disjoint_errors : bool, optional
        Iff True, disjoint error approximations will be allowed.
    """
    dem = stim_circuit.detector_error_model(
        decompose_errors=False, approximate_disjoint_errors=approximate_disjoint_errors
    )

    detector_counter = DetectorCounter()
    DemParser(
        detector_counter,
        lambda detector, detector_offset, coordinate_offset: None,  # noqa: ARG005
    ).parse(dem)

    max_num_detectors_for_graph_like: int = 2
    if detector_counter.max_num_detectors() > max_num_detectors_for_graph_like:
        dem = stim_circuit.detector_error_model(
            decompose_errors=True,
            approximate_disjoint_errors=approximate_disjoint_errors,
        )

    return dem


@no_type_check
def graph_to_json(
    decoding_graph: HyperMultiGraph[AnyEdgeT],
    logicals: Iterable[set[AnyEdgeT]],
    full: bool = False,
) -> str:
    """Represent some graph as an edge list in JSON. Logicals are also included.
    Fields are:

    - edges: List of sorted tuples for all edges
    - edge_weights: list of floats for edge weights
    - logicals: List of sorted tuples for logical edges
    - boundary: Decoding graph boundary. None if it doesn't exist.

    If full flag is True, we additionally serialise:

    - detector_records: dict containing detector metadata
    - edge_record: dict containing edge metadata

    Parameters
    ----------
    decoding_graph :  HyperMultiGraph[AnyEdgeT]
        Graph to represent in JSON. Should have no more than one boundary.
    logicals : Iterable[Set[AnyEdgeT]]
        Logicals to represent in JSON.
    full : bool
        Whether to save full graph information that includes
        detector and edge records.

    Returns
    -------
    str
        JSON formatted string representing the essential graph and logical info
    """
    if len(decoding_graph.boundaries) > 1:
        msg = "JSON graph representation supports maximum one boundary"
        raise ValueError(msg)

    if isinstance(decoding_graph, (DecodingHyperMultiGraph, NXDecodingMultiGraph)):
        edges = [(sorted(edge), edge_id) for edge, edge_id in decoding_graph.edges]
        logicals = [
            [(sorted(edge), edge_id) for edge, edge_id in logical]
            for logical in logicals
        ]
    else:
        edges = [sorted(edge) for edge in decoding_graph.edges]
        logicals = [[sorted(edge) for edge in logical] for logical in logicals]

    edge_weights = [
        decoding_graph.edge_records[edge].weight for edge in decoding_graph.edges
    ]
    boundary = (
        next(iter(decoding_graph.boundaries)) if decoding_graph.boundaries else None
    )

    graph_as_dict = {
        "edges": edges,
        "edge_weights": edge_weights,
        "logicals": logicals,
        "boundary": boundary,
    }
    if full:
        detector_records = {
            key: dict(value) for key, value in decoding_graph.detector_records.items()
        }
        edge_records = []
        for edge in edges:
            record = dict(decoding_graph.edge_records[DecodingEdge(*edge)])
            record.pop("decoding_edge", None)
            edge_records.append(record)
        graph_as_dict["detector_records"] = detector_records
        graph_as_dict["edge_records"] = edge_records

    return json.dumps(graph_as_dict)


def nx_graph_from_json(
    json_str: str,
) -> tuple[NXDecodingGraph, list[frozenset[DecodingEdge]]]:
    """Loads the graph from json string given by graph_to_json.

    Parameters
    ----------
    json_str : str
        JSON string.

    Returns
    -------
    Tuple[NXDecodingGraph, List[FrozenSet[DecodingEdge]]]
        Reconstructed decoding graph and logicals.
    """
    graph_as_dict = json.loads(json_str)
    if (boundary_node := graph_as_dict["boundary"]) is not None:
        boundary = frozenset([boundary_node])
    else:
        boundary = frozenset()
    # collect detector records if defined
    detector_records = None
    if (detector_records_dict := graph_as_dict.get("detector_records")) is not None:
        detector_records = {
            int(key): DetectorRecord.from_dict(value)
            for key, value in detector_records_dict.items()
        }
    # collect edge data if defined
    edge_data: Iterable[DecodingEdge | tuple[DecodingEdge, EdgeRecord]]
    if (edge_records_dict := graph_as_dict.get("edge_records")) is not None:
        edge_records = [EdgeRecord.from_dict(record) for record in edge_records_dict]
        edge_data = [
            (DecodingEdge(*edge), record)
            for edge, record in zip(graph_as_dict["edges"], edge_records, strict=False)
        ]
    else:
        edge_data = [DecodingEdge(*edge) for edge in graph_as_dict["edges"]]
    graph = NXDecodingGraph.from_edge_list(edge_data, detector_records, boundary)
    logicals = [
        frozenset(DecodingEdge(*edge) for edge in logical)
        for logical in graph_as_dict["logicals"]
    ]
    return graph, logicals


def inverse_logical_at_boundary(
    decoding_graph: NXDecodingGraph, logical: set[DecodingEdge]
) -> set[DecodingEdge]:
    """Given a decoding graph and a logical, that is assumed to be along the
    boundary, return another logical constructed of all the edges incident to the
    boundary that are not in the given logical.

    On a surface code for example, if we gave the logical as the edges along one
    side of the surface, this function would return the logical made of the edges
    along the opposite side of the surface.

    Parameters
    ----------
    decoding_graph : NXDecodingGraph
        Decoding graph the logicals are defined over.
    logical : Set[DecodingEdge]
        The logical to get the boundary inverse of.
        If this given logical is not consisting of only edges along the boundary, then
        an exception will be raised.

    Returns
    -------
    Set[DecodingEdge]
        A logical along the boundary, the inverse of the given logical.

    Raises
    ------
    ValueError
        Exception raised if the the logical is not entirely along the boundary of the
        graph. This includes if the logical has an edge not in the graph.
    """
    all_edges_to_boundaries: set[DecodingEdge] = set()
    for boundary in decoding_graph.boundaries:
        all_edges_to_boundaries.update(decoding_graph.incident_edges(boundary))

    inverse_logical_edges = {
        edge for edge in all_edges_to_boundaries if edge not in logical
    }
    if len(inverse_logical_edges | logical) != len(all_edges_to_boundaries):
        msg = (
            "The logical given was not entirely along the boundary, "
            "so the inverse logical is undefined."
        )
        raise ValueError(msg)
    return inverse_logical_edges


def is_logical_along_boundary(
    decoding_graph: HyperMultiGraph, logicals: HyperLogicals
) -> bool:
    """Given a decoding graph and logicals, check if the logicals are along the boundary

    Parameters
    ----------
    decoding_graph : HyperMultiGraph
        The decoding graph that the logicals are defined over.
    logicals : HyperLogicals
        The logicals to check.

    Returns
    -------
    bool
        True if the logicals are along the boundary.
    """
    return all(
        any(
            incident_syndrome_bit in decoding_graph.boundaries
            for incident_syndrome_bit in tuple(edge.vertices)
        )
        for logical in logicals
        for edge in logical
    )


def worst_case_num_detectors(
    decoding_graph: DecodingHyperGraph, target_logical_error: float
) -> int:
    """For a given `decoding_graph` representing error mechanisms, and a
    `target_logical_error`, return the largest number of detection events we would
    expect to observe on the `decoding_graph`.

    Parameters
    ----------
    decoding_graph : DecodingHyperGraph
        Decoding hyper graph, where each edge is an error mechanism with some
        probability of occurrence `p_err`.
    target_logical_error : float

    Returns
    -------
    int
        Largest number of detection events we expect to observe.
    """

    def poisson_pmf(expected, k: int):
        expected = decimal.Decimal(expected)  # For increased precision
        return (expected**k) * decimal.Decimal(-expected).exp() / math.factorial(k)

    expected_num_detectors = sum(
        edge_record.p_err * len(edge)
        for edge, edge_record in decoding_graph.edge_records.items()
    )

    if expected_num_detectors == 0:
        return 0

    cumulative_p = 0
    for num_detectors in range(
        len(decoding_graph.nodes) - len(decoding_graph.boundaries) + 1
    ):
        cumulative_p += poisson_pmf(expected_num_detectors, num_detectors)

        if (1 - cumulative_p) <= target_logical_error:
            break

    return num_detectors


def compute_graph_distance_for_logical(
    decoding_graph: NXDecodingGraph,
    logical: AbstractSet[DecodingEdge],
    weighted: bool = False,
) -> int | float:
    """Computes the distance of the decoding graph given a logical.
    This is done by calculating the shortest path
    between nodes of the logical edges in the graph where all edges of
    that logical have been removed.
    If weighted, returns the length of the shortest non-trivial loop.

    Note: this method is not fully general and will not work well for all
    definitions of the logical! In particular, it fails if the smallest
    weight logical operator has a support on the logical greater than 1.
    I.e. if non-trivial logical loop of edges crosses the logical more
    than once. However, this should be satisfied for most surface codes
    with usual definition of logicals.
    It is likely possible to do this more generally on a matching graph
    by trying to find a shortest path from every point to itself (using
    a modified Dijkstra) that contains the odd number of logical crossings.

    Parameters
    ----------
    decoding_graph : NXDecodingGraph
        Decoding graph to calculate the distance on.
    logical : AbstractSet[DecodingEdge]
        Reference logical.
    weighted : bool
        Whether to calculate the distance in terms of number of edges
        or the weight.
    Returns
    -------
    Union[int, float]
        Distance of the logical. If weighted, returns the number of edges
        (int), otherwise returns the length of the shortest non-trivial loop (float)
    """
    decoding_graph_no_logicals = nx.subgraph_view(
        decoding_graph.graph, filter_edge=lambda u, v: DecodingEdge(u, v) not in logical
    )
    weight = "weight" if weighted else None
    # get the minimum weight of a logical loop by taking the minimal
    # path around a logical edge and adding the weight of that edge
    if weighted:
        edge_weights = np.array(
            [decoding_graph.edge_records[edge].weight for edge in logical]
        )
    else:
        edge_weights = np.ones(len(logical), dtype=int)
    path_lengths = np.array(
        [
            nx.dijkstra_path_length(
                decoding_graph_no_logicals, edge.first, edge.second, weight=weight
            )
            for edge in logical
        ]
    )
    return (path_lengths + edge_weights).min().item()


def compute_graph_distance(
    decoding_graph: NXDecodingGraph, logicals: NXLogicals, weighted: bool = False
) -> int | float:
    """Calculates the minimum distance for all logicals.
    See `compute_graph_distance_for_logical`.
    Decoding graph to calculate the distance on.

    logicals : NXLogicals
        Reference logicals.
    weighted : bool
        Whether to calculate the distance in terms of number of edges
        or the weight.

    Note: this method is not fully general and will not work well for all
    definitions of the logical! In particular, it fails if the smallest
    weight logical operator has a support on the logical greater than 1.
    I.e. if non-trivial logical loop of edges crosses the logical more
    than once. However, this should be satisfied for most surface codes
    with usual definition of logicals.
    It is likely possible to do this more generally on a matching graph
    by trying to find a shortest path from every point to itself (using
    a modified Dijkstra) that contains the odd number of logical crossings.

    Returns
    -------
    Union[int, float]
        Decoding graph distance. Returns float if weighted is True.
    """
    return min(
        compute_graph_distance_for_logical(decoding_graph, logical, weighted=weighted)
        for logical in logicals
    )


def unweight_graph(decoding_graph: NXDecodingGraph):
    """Sets all weights in the given graph to 1 and error probabilities to 1/(1+e)
    so that they correspond to the weights.

    Parameters
    ----------
    decoding_graph : NXDecodingGraph
        Original decoding to unweight
    """
    for edge in decoding_graph.edges:
        decoding_graph.edge_records[edge].p_err = 1.0 / (1.0 + math.e)
        # set the weight as well just to be sure there are no floating point errors
        decoding_graph.edge_records[edge]["weight"] = 1.0



# (c) Copyright Riverlane 2020-2025.
"""Datastructures for data qubits."""

from __future__ import annotations

import math
from collections import Counter, UserDict
from collections.abc import (
    Collection,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
)
from collections.abc import (
    Set as AbstractSet,
)
from functools import cached_property
from itertools import chain
from typing import Any, Generic, TypeVar

from deltakit_core.decoding_graphs._syndromes import (
    Bit,
    DetectorRecord,
    OrderedSyndrome,
)


class EdgeRecord(UserDict):
    """Dictionary for recording information about an edge.
    String attributes for arbitrary values. Error probability
    given as special data that is always defined.

    Parameters
    ----------
    p_err : float, optional
        Probability of the error mechanism, by default 0.0.
    """

    def __init__(self, p_err: float = 0.0, **kwargs):
        super().__init__(p_err=p_err, **kwargs)
        self.data["weight"] = self.weight

    @property
    def weight(self) -> float:
        """Log-likelihood weight of the error mechanism."""
        if self.p_err == 0:
            return math.inf
        if self.p_err == 1:
            return -math.inf
        if not 0 < self.p_err < 1:
            msg = f"Edge weight undefined for error probability {self.p_err}"
            raise ValueError(msg)
        return math.log((1 - self.p_err) / self.p_err)

    @property
    def p_err(self) -> float:
        """Probability of the error mechanism occurring."""
        return self.data["p_err"]

    @p_err.setter
    def p_err(self, value: float):
        self.data["p_err"] = value
        self.data["weight"] = self.weight

    @classmethod
    def from_dict(cls, property_dict: dict[str, Any]) -> EdgeRecord:
        """Create a EdgeRecord from a given property dict of optional values.

        Parameters
        ----------
        property_dict : Dict[str, Any]
            Any optional properties such as p_err.

        Returns
        -------
        EdgeRecord
        """
        record = EdgeRecord(p_err=property_dict.get("p_err", 0.0))
        record.update(property_dict)
        return record

    @classmethod
    def from_loglikelihood(cls, weight: float) -> EdgeRecord:
        """Create a EdgeRecord from a loglikelihood weight rather than a p_err.

        Parameters
        ----------
        weight: float
            Edge weight as a loglikelihood, defined as w=ln((1-p)/p)
            where w is the weight, p is error probability and
            ln is the natural log.

        Returns
        -------
        EdgeRecord
        """
        return EdgeRecord(p_err=1 / (1 + math.exp(weight)))


class DecodingHyperEdge(Collection[int]):
    """Representation of an immutable decoding edge on a hyper-graph.
    This is a set of syndrome bits.

    Parameters
    ----------
    vertices : Iterable[int]
        Syndrome bits connected by this edge.
    """

    def __init__(self, vertices: Iterable[int]):
        self._vertices = frozenset(vertices)
        self._hash = hash(self.vertices)

    @property
    def vertices(self) -> frozenset[int]:
        """Vertices in this edge."""
        return self._vertices

    def to_decoding_edge(self, boundary: int | None = None) -> DecodingEdge:
        """Cast this edge into a decoding edge if possible."""
        degree_target = 2
        if (degree := len(self.vertices)) == degree_target:
            return DecodingEdge(*self.vertices)
        if degree == 1:
            if boundary is None:
                msg = f"Boundary vertex is required for edge: {self} of degree one."
                raise ValueError(msg)
            return DecodingEdge(next(iter(self.vertices)), boundary)
        msg = f"Cannot cast edge: {self} of degree {degree} to decoding edge."
        raise ValueError(msg)

    def __repr__(self) -> str:
        return str(tuple(self.vertices))

    def __hash__(self):
        return self._hash

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, DecodingHyperEdge)
            and self._hash == other._hash
            and self.vertices == other.vertices
        )

    def __iter__(self) -> Iterator[int]:
        return self.vertices.__iter__()

    def __len__(self) -> int:
        return len(self.vertices)

    def __contains__(self, x: object) -> bool:
        return x in self.vertices


class DecodingEdge(DecodingHyperEdge):
    """Representation of an immutable decoding edge on a decoding graph, limited to
    standard graphs where edges connect two distinct detectors.

    Parameters
    ----------
    first_detector : int
        First detector this edge is incident to.
    second_detector : int
        Second detector this edge is incident to.
    """

    def __init__(self, first_detector: int, second_detector: int):
        if first_detector == second_detector:
            msg = (
                f"Invalid DecodingEdge between detectors {first_detector} "
                f"and {second_detector}."
            )
            raise ValueError(msg)
        super().__init__((first_detector, second_detector))

    @property
    def first(self) -> int:
        """Return the first vertex in this edge, order is arbitrary."""
        return next(iter(self._vertices))

    @property
    def second(self) -> int:
        """Return the second vertex in this edge, order is arbitrary."""
        return tuple(self._vertices)[1]

    def is_timelike(self, detector_records: Mapping[int, DetectorRecord]) -> bool:
        """Return True if this decoding edge is between the same data qubit at different
        time steps.
        """
        return (
            detector_records[self.first].time != detector_records[self.second].time
        ) and (
            detector_records[self.first].spatial_coord
            == detector_records[self.second].spatial_coord
        )

    def is_spacelike(self, detector_records: Mapping[int, DetectorRecord]) -> bool:
        """Return True if this decoding edge is between the same data qubit at different
        time steps.
        """
        return (
            detector_records[self.first].time == detector_records[self.second].time
        ) and (
            detector_records[self.first].spatial_coord
            != detector_records[self.second].spatial_coord
        )

    def is_hooklike(self, detector_records: Mapping[int, DetectorRecord]) -> bool:
        """Return True if this decoding edge is between the same data qubit at different
        time steps.
        """
        return (
            detector_records[self.first].time != detector_records[self.second].time
        ) and (
            detector_records[self.first].spatial_coord
            != detector_records[self.second].spatial_coord
        )


EdgeT = TypeVar("EdgeT", bound=DecodingHyperEdge)


class OrderedDecodingEdges(Sequence[EdgeT], AbstractSet[EdgeT], Generic[EdgeT]):
    """Immutable ordered mod 2 set of decoding edges.

    All decoding edges are thought to be on the same decoding graph.
    I.e. if the z and x decoding graphs are completely separable, then the z and x
    errors will be represented in two different `OrderedDecodingEdges` objects.

    Parameters
    ----------
    decoding_edges : Optional[Iterable[EdgeT]], optional
        Iterable of edges to store, in the given order, by default None. Duplicate
        items are removed if they are present an even number of times. Duplicate items
        are reduced to a single occurrence if they are present an odd number of times.
    mod_2_filter : bool, optional
        False disables the modulo 2 filter, in which case the input
        edges are treated as an ordered set, by default True.
    """

    def __init__(
        self,
        decoding_edges: Iterable[EdgeT] | None = None,
        mod_2_filter: bool = True,
    ):
        _decoding_edges = [] if decoding_edges is None else decoding_edges
        if mod_2_filter:
            self._decoding_edges = self._mod_2_filter(_decoding_edges)
        else:
            self._decoding_edges = dict.fromkeys(_decoding_edges)

    @staticmethod
    def _mod_2_filter(decoding_edges: Iterable[EdgeT]) -> dict[EdgeT, None]:
        edge_counts = Counter(decoding_edges)
        return dict.fromkeys(
            [edge for edge, count in edge_counts.items() if count % 2 == 1]
        )

    def append(self, other: OrderedDecodingEdges, mod_2_filter: bool = False):
        """Appeds edges from another set of OrderedDecodingEdges."""
        # pylint: disable=protected-access
        if mod_2_filter:
            self._decoding_edges = self._mod_2_filter(chain(self, other))
        else:
            self._decoding_edges.update(other._decoding_edges)

    @cached_property
    def _as_tuple(self) -> tuple[EdgeT, ...]:
        """Defined to create immutable object to hash, and to make `__getitem__` a O(1)
        method. Exists as a property to avoid duplication of data in core member
        attributes, and to have this be created only when needed.
        """
        return tuple(self._decoding_edges)

    def __add__(self, other: OrderedDecodingEdges) -> OrderedDecodingEdges:
        result = OrderedDecodingEdges(self._decoding_edges.keys(), mod_2_filter=False)
        result.append(other)
        return result

    def __len__(self) -> int:
        return len(self._decoding_edges)

    def __getitem__(self, index):
        return self._as_tuple.__getitem__(index)

    def __contains__(self, x: object) -> bool:
        return self._decoding_edges.__contains__(x)

    def __iter__(self) -> Iterator[EdgeT]:
        return self._decoding_edges.__iter__()

    def __repr__(self) -> str:
        return f"[{', '.join(map(str, self))}]"

    def __hash__(self) -> int:
        return hash(self._as_tuple)

    def __eq__(self, other: object) -> bool:
        # Perhaps allow other sequences to be comparable also?
        return (
            isinstance(other, OrderedDecodingEdges)
            and self._as_tuple == other._as_tuple
        )

    def as_bitstring(self, edges: Sequence[EdgeT]) -> list[Bit]:
        """Convert given edges to a bitstring representation of `len(edges)` bits
        from condition of each edge inside  _decoding_edges.

        Parameters
        ----------
        edges : Sequence[EdgeT]
            Edges to be converted to bitstring .

        Returns
        -------
        List[Bit]
            Bitstring of whether each edge is in `_decoding_edges`.
        """
        return [1 if edge in self else 0 for edge in edges]

    @classmethod
    def from_syndrome_indices(
        cls, indices: Iterable[tuple[int, int]]
    ) -> OrderedDecodingEdges[DecodingEdge]:
        """Given a list of pairs of syndrome indices, construct the corresponding
        decoding edges and return in an `OrderedDecodingEdges` collection.

        Parameters
        ----------
        indices : Iterable[Tuple[int, int]]
            Indices into the syndrome sequence.

        Returns
        -------
        OrderedDecodingEdges
        """
        return OrderedDecodingEdges(DecodingEdge(u, v) for u, v in indices)


def errors_to_syndrome(
    decoding_edges: Iterable[EdgeT], boundaries: AbstractSet[int] = frozenset()
) -> OrderedSyndrome:
    """Take decoding edges and convert to an ordered syndrome that would be observed
    if all decoding edges were to be active.

    Parameters
    ----------
    decoding_edges : Iterable[EdgeT]
        Decoding edges to convert.
    boundaries : AbstractSet[int], optional
        Set of detectors that are virtual boundary detectors and should hence not show
        in a syndrome. By default None.

    Returns
    -------
    OrderedSyndrome
        Resultant ordered syndrome.
    """
    return OrderedSyndrome(
        symptom
        for symptom in chain.from_iterable(decoding_edges)
        if symptom not in boundaries
    )



# (c) Copyright Riverlane 2020-2025.
"""Module for utilities on annotating hypergraphs with window ids."""

from collections import defaultdict
from collections.abc import Callable
from itertools import combinations

from networkx import DiGraph
from networkx.algorithms import simple_cycles

from deltakit_core.decoding_graphs._data_qubits import DecodingHyperEdge
from deltakit_core.decoding_graphs._decoding_graph import (
    DecodingHyperGraph,
    HyperLogicals,
)

CoordinateToWindowId = Callable[[tuple[float, ...]], int]


def annotate_edges_with_window_ids(
    hypergraph: DecodingHyperGraph,
    coordinate_to_window_id: CoordinateToWindowId,
    clear_edges: bool = True,
) -> None:
    """Annotate edges with window_ids derived from node coordinates.

    For each detector, compute a window_id using the provided
    `coordinate_to_window_id` function (based on the node's full_coord).
    Each edge's `window_id` is set to the minimum window_id of its adjacent
    nodes. Optionally clears any existing edge `window_id` annotations first.

    Using the minimum detector enforces that windows flow from lower to higher
    `window ids`, which produces an acyclic transfer graph when deriving inter-window
    dependencies.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Hypergraph whose edges will be annotated.
    coordinate_to_window_id : Callable[[tuple[float, ...]], int]
        Function mapping a detector's full_coord (including time) to a window id.
    clear_edges : bool
        If True, clear existing edge window_ids before annotating. Default True.
    """
    if clear_edges:
        for edge_record in hypergraph.edge_records.values():
            edge_record.pop("window_id", None)

    node_to_window_id = {
        det_id: coordinate_to_window_id(det_record.full_coord)
        for det_id, det_record in hypergraph.detector_records.items()
    }

    for edge, edge_record in hypergraph.edge_records.items():
        edge_record["window_id"] = min(node_to_window_id[det] for det in edge)


def get_window_id_transfer_graph(hypergraph: DecodingHyperGraph) -> DiGraph:
    """Given a decoding hypergraph with `window_id` fields in edge records,
    construct a transfer dependency graph. Transfers are unidirectional
    and always go from lower to higher `window_id`.

    Two window_ids are dependent if there are adjacent edges (sharing the same node)
    with those window_ids.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Global hypergraph where all edges have an associated
        `window_id`.

    Returns
    -------
    DiGraph
        Directed transfer hypergraph whose nodes are the unique
        window_ids of the given hypergraph.
    """
    transfer_graph: DiGraph = DiGraph()
    for node in hypergraph.nodes:
        shared_window_ids = {
            hypergraph.edge_records[edge]["window_id"]
            for edge in hypergraph.incident_edges(node)
        }
        for lbl1, lbl2 in combinations(shared_window_ids, 2):
            src, sink = (lbl2, lbl1) if lbl2 < lbl1 else (lbl1, lbl2)
            transfer_graph.add_edge(src, sink)
    # sanity check that there are no cycles
    if len(list(simple_cycles(transfer_graph))) != 0:
        msg = "Transfer hypergraph cannot have cycles!"
        raise ValueError(msg)
    return transfer_graph


def get_unique_window_ids(hypergraph: DecodingHyperGraph) -> set[int]:
    """Return the set of unique edge `window_id`s in the hypergraph.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Global hypergraph where all edges have an associated `window_id`.

    Returns
    -------
    set[int]
        set of unique edge `window_id`s.
    """
    return {
        edge_record["window_id"] for edge_record in hypergraph.edge_records.values()
    }


def get_unique_logical_window_ids(
    hypergraph: DecodingHyperGraph, logicals: HyperLogicals
) -> set[int]:
    """Gets the unique edge `window_id`s across the provided logicals (edge sets).

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Global hypergraph where all logical edges have an associated window_id.
    logicals : HyperLogicals
        Global logicals to get the window_ids for.

    Returns
    -------
    set[int]
        Unique edge `window_id`s present among logical edges.
    """
    return {
        hypergraph.edge_records[edge]["window_id"]
        for logical in logicals
        for edge in logical
    }


def separate_edges_per_window_id(
    graph: DecodingHyperGraph,
    enforce_window_ids: bool = True,
) -> dict[int, list[DecodingHyperEdge]]:
    """Separates labelled graph edges into lists keyed by ``window_id``.

    Parameters
    ----------
    graph : DecodingHyperGraph
        Hypergraph whose edges are (partially) annotated with ``window_id``.
    enforce_window_ids : bool, optional
        If True, raises ``ValueError`` if any edge lacks a ``window_id`` label.

    Returns
    -------
    dict[int, list[DecodingHyperEdge]]
        Mapping from window id to list of edges with that label.
    """
    windows_edges: dict[int, list[DecodingHyperEdge]] = defaultdict(list)
    unlabelled: list[DecodingHyperEdge] = []
    for edge, edge_record in graph.edge_records.items():
        if (label := edge_record.get("window_id")) is not None:
            windows_edges[label].append(edge)
        else:
            unlabelled.append(edge)
    if enforce_window_ids and unlabelled:
        msg = f"Found {len(unlabelled)} unlabelled edges when enforce_window_ids=True"
        raise ValueError(msg)
    return windows_edges


def separate_logicals_by_window_id(
    hypergraph: DecodingHyperGraph, logicals: HyperLogicals
) -> dict[int, HyperLogicals]:
    """Separate logicals by edge `window_id`.

    Returns a dictionary mapping window_id to a list of logicals (each logical is a
    frozenset of edges) that have that window_id. Keys are in ascending order.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Global hypergraph where all logical edges have an associated window_id.
    logicals : HyperLogicals
        Global logicals.

    Returns
    -------
    dict[int, list[list[DecodingHyperEdge]]]
        dictionary from logical window_ids to the list of lists of edges containing
        logicals that have that window_id.
    """
    num_logicals = len(logicals)
    all_window_ids = get_unique_logical_window_ids(hypergraph, logicals)
    logicals_by_window_id: dict[int, list[list[DecodingHyperEdge]]] = {
        lbl: [[] for _ in range(num_logicals)] for lbl in sorted(all_window_ids)
    }
    for i, logical in enumerate(logicals):
        for edge in logical:
            window_id = hypergraph.edge_records[edge]["window_id"]
            logicals_by_window_id[window_id][i].append(edge)
    return {
        key: [frozenset(log) for log in logicals]
        for key, logicals in logicals_by_window_id.items()
    }



# (c) Copyright Riverlane 2020-2025.
"""Datastructures for decoding graphs."""

from __future__ import annotations

import math
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import (
    Callable,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
)
from collections.abc import (
    Set as AbstractSet,
)
from dataclasses import dataclass
from functools import cached_property
from itertools import chain, tee
from typing import ClassVar, Generic, TypeGuard, TypeVar, cast

import networkx as nx
import numpy as np
import numpy.typing as npt

from deltakit_core.decoding_graphs._data_qubits import (
    DecodingEdge,
    DecodingHyperEdge,
    EdgeRecord,
)
from deltakit_core.decoding_graphs._syndromes import DetectorRecord, OrderedSyndrome


def _is_decoding_edge_and_edge_record(
    data: DecodingEdge | tuple[int, ...] | tuple[DecodingEdge, EdgeRecord],
) -> TypeGuard[tuple[DecodingEdge, EdgeRecord]]:
    return (
        not isinstance(data, DecodingEdge)
        and isinstance(data[0], DecodingEdge)
        and isinstance(data[-1], EdgeRecord)
    )


def _is_decoding_hyper_edge_and_edge_record(
    data: DecodingHyperEdge | tuple[int, ...] | tuple[DecodingHyperEdge, EdgeRecord],
) -> TypeGuard[tuple[DecodingHyperEdge, EdgeRecord]]:
    return (
        not isinstance(data, DecodingHyperEdge)
        and isinstance(data[0], DecodingHyperEdge)
        and isinstance(data[-1], EdgeRecord)
    )


def _is_tuple_of_ints(
    data: DecodingHyperEdge | tuple[int, ...] | tuple[DecodingHyperEdge, EdgeRecord],
) -> TypeGuard[tuple[int, ...]]:
    return not isinstance(data, DecodingHyperEdge) and all(
        isinstance(v, int) for v in data
    )


DecodingGraphT = TypeVar("DecodingGraphT")
AnyEdgeT = TypeVar("AnyEdgeT")


def convolve_probabilities(edge_records: Iterable[EdgeRecord]) -> EdgeRecord:
    """Computes the probability that an odd number of independent events
    with the given probabilities occur.

    Parameters
    ----------
    edge_records : Iterable[EdgeRecord]
        All of the edge records associated with a particular edge.

    Returns
    -------
    EdgeRecord
        A single edge record which is the product of the individual
        probabilities of each edge record.
    """
    combined_probabilities = math.prod(1 - 2 * record.p_err for record in edge_records)
    return EdgeRecord((1 - combined_probabilities) / 2)


class HyperMultiGraph(ABC, Generic[AnyEdgeT]):
    """Class for abstract immutable multigraph with hyperedges. Any set of
    vertices can be connected by one or more edges in the multi-hypergraph,
    forming the basis of all other graph objects."""

    @property
    @abstractmethod
    def edges(self) -> Sequence[AnyEdgeT]:
        """Return all edges of this graph."""

    @property
    @abstractmethod
    def nodes(self) -> Sequence[int]:
        """Return all nodes of this graph."""

    @property
    @abstractmethod
    def boundary_edges(self) -> Iterable[AnyEdgeT]:
        """Fetch the edges that are touching the boundary of the graph."""

    @abstractmethod
    def get_edges(self, *detectors: int) -> Iterable[AnyEdgeT]:
        """Generator for the edges defined by the given set of detectors."""

    @abstractmethod
    def incident_edges(self, detector: int) -> Iterator[AnyEdgeT]:
        """Iterator for the incident edges connected to a given detector."""

    @abstractmethod
    def neighbors(self, detector: int) -> Iterator[int]:
        """Iterator for the neighbours of a given detector."""

    @property
    @abstractmethod
    def edge_records(self) -> Mapping[AnyEdgeT, EdgeRecord]:
        """Return mapping of edge records, keyed by edge."""

    @property
    @abstractmethod
    def detector_records(self) -> Mapping[int, DetectorRecord]:
        """Return mapping of detector records, keyed by detector."""

    @property
    def boundaries(self) -> frozenset[int]:
        """Return all detectors in this graph that are labelled as boundaries."""
        return frozenset()

    @abstractmethod
    def detector_is_boundary(self, detector: int) -> bool:
        """Return True if given detector is a boundary, False otherwise."""

    @abstractmethod
    def to_parity_check_matrix(self) -> npt.NDArray[np.uint8]:
        """Convert the hypergraph to a parity check matrix  of size
        `(len(nodes), len(edges))`. Each column represents an edge and
        the non-zero entries in that column represent detectors in the edge.

        The nodes in the graph must be contiguous to construct the matrix.
        """

    @abstractmethod
    def error_to_syndrome(self, edges: Iterable[AnyEdgeT]) -> OrderedSyndrome:
        """Assume an error has happened on the given edges and return the
        corresponding syndrome.

        Parameters
        ----------
        edges : Iterable[AnyEdgeT]
            Edges which had errors.

        Returns
        -------
        OrderedSyndrome
        """


class DecodingHyperMultiGraph(HyperMultiGraph[tuple[DecodingHyperEdge, int]]):
    """Representation of a decoding multigraph with hyper-edges.

    Parameters
    ----------
    multi_hyper_edges: Iterable[Tuple[DecodingHyperEdge, int]]
        Edges that form this graph.
    detector_records : Optional[Dict[int, DetectorRecord]], optional
        Optional mapping of detector indices to detector records, by default None.
    """

    def __init__(
        self,
        edge_data: Iterable[
            DecodingHyperEdge | tuple[int, ...] | tuple[DecodingHyperEdge, EdgeRecord]
        ],
        detector_records: dict[int, DetectorRecord] | None = None,
    ):
        edge_records: dict[tuple[DecodingHyperEdge, int], EdgeRecord] = {}
        edges: list[tuple[DecodingHyperEdge, int]] = []
        edge_count: dict[DecodingHyperEdge, int] = defaultdict(int)

        for data in edge_data:
            if isinstance(data, DecodingHyperEdge):
                edge = data
                edge_record = EdgeRecord()
            elif _is_decoding_hyper_edge_and_edge_record(data):
                edge, edge_record = data
            elif _is_tuple_of_ints(data):
                edge = DecodingHyperEdge(data)
                edge_record = EdgeRecord()
            else:
                msg = f"Invalid edge data {data}"
                raise ValueError(msg)
            edge_id = edge_count[edge]
            edge_count[edge] += 1
            edge_records[(edge, edge_id)] = edge_record
            edges.append((edge, edge_id))

        self._edges = edges
        if len(set(edges)) != len(edges):
            msg = "Each edge must have a unique integer identifier"
            raise ValueError(msg)
        self._detector_records = {} if detector_records is None else detector_records
        self._edge_records = edge_records

    @cached_property
    def _nodes_in_edges(self):
        return {node for edge, _ in self._edges for node in edge}

    @property
    def detector_records(self) -> dict[int, DetectorRecord]:
        return self._detector_records

    @property
    def edge_records(self) -> dict[tuple[DecodingHyperEdge, int], EdgeRecord]:
        return self._edge_records

    @property
    def edges(self) -> list[tuple[DecodingHyperEdge, int]]:
        return self._edges

    @cached_property
    def nodes(self) -> list[int]:
        return sorted(self._nodes_in_edges.union(set(self._detector_records.keys())))

    @cached_property
    def boundary_edges(self) -> list[tuple[DecodingHyperEdge, int]]:
        return [(edge, edge_id) for edge, edge_id in self.edges if len(edge) == 1]

    @cached_property
    def _nodes_to_full_edge(
        self,
    ) -> defaultdict[frozenset[int], list[tuple[DecodingHyperEdge, int]]]:
        nodes_to_full_edge = defaultdict(list)
        for edge, edge_id in self.edges:
            nodes_to_full_edge[edge.vertices].append((edge, edge_id))
        return nodes_to_full_edge

    @cached_property
    def _node_to_incident_edges(
        self,
    ) -> defaultdict[int, list[tuple[DecodingHyperEdge, int]]]:
        detector_to_edges = defaultdict(list)
        for edge, edge_id in self.edges:
            for node in edge.vertices:
                detector_to_edges[node].append((edge, edge_id))
        return detector_to_edges

    def get_edges(self, *detectors: int) -> list[tuple[DecodingHyperEdge, int]]:
        """Fetch the edges connecting a sequence of detectors."""
        return self._nodes_to_full_edge[frozenset(detectors)]

    def incident_edges(self, detector: int) -> Iterator[tuple[DecodingHyperEdge, int]]:
        yield from self._node_to_incident_edges[detector]

    def neighbors(self, detector: int) -> Iterator[int]:
        incident_edges = self.incident_edges(detector)
        neighbour_nodes = {
            node
            for edge, _ in incident_edges
            for node in edge.vertices
            if node != detector
        }
        yield from neighbour_nodes

    def detector_is_boundary(self, detector: int) -> bool:  # noqa: ARG002
        return False

    def to_parity_check_matrix(self) -> npt.NDArray[np.uint8]:
        check_matrix = np.zeros((len(self.nodes), len(self.edges)), dtype=np.uint8)
        for edge_index, (edge, _) in enumerate(self.edges):
            check_matrix[tuple(edge), (edge_index,)] = True
        return check_matrix

    def with_multi_edges_merged(
        self,
        combine_edge_records: Callable[
            [Iterable[EdgeRecord]], EdgeRecord
        ] = convolve_probabilities,
    ) -> DecodingHyperGraph:
        """Create a `DecodingHyperGraph` instance with the same nodes but with
        only single edges, merging the multi-edge edge records to a single edge
        record.

        Parameters
        ----------
        combine_edge_records : Callable[[Iterable[EdgeRecord]], EdgeRecord], optional
            A function which combines all the edge records for a given edge,
            by default convolve_probabilities which applies the generalised
            `p * (1 - q) + q * (1 - p)`.

        Returns
        -------
        DecodingHyperGraph
            A hypergraph with the same properties as this graph but with all
            multi-edges merged to a single edge.
        """
        grouped_edge_records = defaultdict(list)
        for (edge, _), edge_record in self.edge_records.items():
            grouped_edge_records[edge].append(edge_record)
        return DecodingHyperGraph(
            edge_data=(
                (edge, combine_edge_records(records))
                for edge, records in grouped_edge_records.items()
            ),
            detector_records=self.detector_records,
        )

    def error_to_syndrome(
        self, edges: Iterable[tuple[DecodingHyperEdge, int]]
    ) -> OrderedSyndrome:
        return OrderedSyndrome(
            symptom for symptom in chain.from_iterable(edge for edge, _ in edges)
        )


HyperLogicals = Sequence[AbstractSet[DecodingHyperEdge]]


class DecodingHyperGraph(HyperMultiGraph[DecodingHyperEdge]):
    """Representation of a decoding hypergraph, built to handle Stim-based noise
    sources. Each hyperedge in the graph has a weight between 0 and 1 (inclusive).

    Parameters
    ----------
    hyper_edges : Iterable[DecodingHyperEdge]
        Edges that form this graph.
    detector_records : Optional[Dict[int, DetectorRecord]], optional
        Optional mapping of detector indices to detector records, by default None.
    edge_records : Optional[Dict[DecodingHyperEdge, EdgeRecord]], optional
        Optional mapping of edges to edge records, by default None.
    """

    def __init__(
        self,
        edge_data: Iterable[
            DecodingHyperEdge | tuple[int, ...] | tuple[DecodingHyperEdge, EdgeRecord]
        ],
        detector_records: dict[int, DetectorRecord] | None = None,
    ):
        edge_records: dict[DecodingHyperEdge, EdgeRecord] = {}
        edges: list[DecodingHyperEdge] = []
        for data in edge_data:
            if _is_decoding_hyper_edge_and_edge_record(data):
                edge, edge_record = data
                edge_records[edge] = edge_record
            elif isinstance(data, DecodingHyperEdge):
                edge = data
                edge_records[edge] = EdgeRecord()
            elif _is_tuple_of_ints(data):
                edge = DecodingHyperEdge(data)
                edge_records[edge] = EdgeRecord()
            else:
                msg = f"Invalid edge data {data}"
                raise ValueError(msg)
            edges.append(edge)

        self._edges = edges
        if len(set(edges)) != len(edges):
            msg = "Each edge must have a unique integer identifier"
            raise ValueError(msg)
        self._detector_records = {} if detector_records is None else detector_records
        self._edge_records = edge_records
        self._nodes_in_detector_records = set(self._detector_records.keys())

    @cached_property
    def _nodes_in_edges(self):
        return set(chain.from_iterable(self._edges))

    @property
    def detector_records(self) -> dict[int, DetectorRecord]:
        return self._detector_records

    @property
    def edge_records(self) -> dict[DecodingHyperEdge, EdgeRecord]:
        return self._edge_records

    @cached_property
    def _node_to_incident_edges(self) -> defaultdict[int, list[DecodingHyperEdge]]:
        detector_to_edges = defaultdict(list)
        for edge in self.edges:
            for node in edge.vertices:
                detector_to_edges[node].append(edge)
        return detector_to_edges

    @cached_property
    def edges(self) -> list[DecodingHyperEdge]:
        return self._edges

    @cached_property
    def nodes(self) -> list[int]:
        return sorted(self._nodes_in_edges.union(self._nodes_in_detector_records))

    @cached_property
    def boundary_edges(self) -> list[DecodingHyperEdge]:
        return [edge for edge in self.edges if len(edge) == 1]

    def get_edges(self, *detectors: int) -> list[DecodingHyperEdge]:
        return (
            [edge]
            if (edge := DecodingHyperEdge(detectors)) in self.edge_records
            else []
        )

    def get_edge(self, *detectors: int) -> DecodingHyperEdge:
        """Returns the unique edge in the graph corresponding to the given set
        of detector indices.

        Raises
        ------
        ValueError
            If no edge in the graph corresponds to the given detectors.
        """
        if (edge := DecodingHyperEdge(detectors)) in self.edges:
            return edge
        msg = f"Detectors {detectors} do not belong to any edge in the graph."
        raise ValueError(msg)

    def get_edge_record(self, *detectors: int) -> EdgeRecord:
        """Fetch the edge record for a sequence of detectors."""
        return self.edge_records[self.get_edge(*detectors)]

    def incident_edges(self, detector: int) -> Iterator[DecodingHyperEdge]:
        yield from self._node_to_incident_edges[detector]

    def neighbors(self, detector: int) -> Iterator[int]:
        incident_edges = self.incident_edges(detector)
        neighbour_nodes = {
            node
            for edge in incident_edges
            for node in edge.vertices
            if node != detector
        }
        yield from neighbour_nodes

    def to_nx_decoding_graph(self) -> NXDecodingGraph:
        """Lower hypergraph to NXDecodingGraph, on the condition that the
        hypergraph does not contain any hyperedges.
        """
        if not self.nodes:
            return NXDecodingGraph.from_edge_list([])
        possible_boundary = max(self.nodes) + 1
        normal_edges: list[tuple[DecodingEdge, EdgeRecord]] = [
            (edge.to_decoding_edge(possible_boundary), self.edge_records[edge])
            for edge in self._edges
        ]
        return NXDecodingGraph.from_edge_list(
            normal_edges,
            self._detector_records,
            boundaries=(
                [possible_boundary]
                if any(possible_boundary in edge for edge, _ in normal_edges)
                else []
            ),
        )

    def to_parity_check_matrix(self) -> npt.NDArray[np.uint8]:
        check_matrix = np.zeros((len(self.nodes), len(self.edges)), dtype=np.uint8)
        for edge_index, edge in enumerate(self.edges):
            check_matrix[tuple(edge), (edge_index,)] = True
        return check_matrix

    def detector_is_boundary(self, detector: int) -> bool:  # noqa: ARG002
        return False

    def error_to_syndrome(self, edges: Iterable[DecodingHyperEdge]) -> OrderedSyndrome:
        return OrderedSyndrome(symptom for symptom in chain.from_iterable(edges))

    def get_relevant_nodes(
        self, logicals: Iterable[set[DecodingHyperEdge]]
    ) -> set[int]:
        """Return nodes that have a hyperedge with a path to logical.

        Hyperedges to boundaries are ignored if applicable.
        """
        union_find: nx.utils.UnionFind = nx.utils.UnionFind()
        for edge in self.edges:
            union_find.union(*edge)

        relevant_nodes: set[int] = set()
        components: list[set[int]] = list(union_find.to_sets())
        for logical in logicals:
            relevant_nodes.update(chain.from_iterable(logical))
            for component in components:
                if any(component.intersection(edge) for edge in logical):
                    relevant_nodes.update(component)
        return relevant_nodes


@dataclass(frozen=True)
class DecodingCode:
    """Convenience class encapsulating the annotated hypergraph and logicals."""

    hypergraph: DecodingHyperGraph
    logicals: HyperLogicals

    def to_nx_code(self) -> NXCode:
        """Lower hypergraph to graph, if there are no hyper edges."""
        graph = self.hypergraph.to_nx_decoding_graph()
        return NXCode(
            graph=graph,
            logicals=[
                {
                    edge.to_decoding_edge(next(iter(graph.boundaries)))
                    for edge in edge_set
                }
                for edge_set in self.logicals
            ],
        )


class _QECNX(nx.Graph):
    """Custom nx graph subclass that uses special dictionary classes for attributes."""

    node_attr_dict_factory = DetectorRecord
    edge_attr_dict_factory = EdgeRecord


class _QECNXMG(nx.MultiGraph):
    """Custom nx.MultiGraph subclass that uses special dictionary classes for
    attributes."""

    node_attr_dict_factory = DetectorRecord
    edge_attr_dict_factory = EdgeRecord


NXGraphT = TypeVar("NXGraphT", bound=nx.Graph)


class NXGraph(HyperMultiGraph[AnyEdgeT], Generic[NXGraphT, AnyEdgeT]):
    """Class for all NetworkX-based graph methods that are shared between
    NXDecodingGraph and NXDecodingMultiGraph
    """

    def __init__(self, graph: NXGraphT, boundaries: Iterable[int] = frozenset()):
        self._graph = nx.freeze(graph)
        self._boundaries = frozenset(boundaries)
        if any(boundary not in self.nodes for boundary in self.boundaries):
            msg = f"Boundaries {boundaries} are not in nodes {self.nodes}."
            raise ValueError(msg)
        max_index = max(self.nodes, default=0)
        self._boundaries_lookup = tuple(
            i in self.boundaries for i in range(max_index + 1)
        )

    def __getstate__(self):
        inner_state = self.__dict__.copy()
        if "no_boundary_view" in inner_state:
            del inner_state["no_boundary_view"]
        return inner_state

    @property
    def graph(self) -> NXGraphT:
        """Get an immutable view of the whole NetworkX graph."""
        return self._graph

    @cached_property
    def no_boundary_view(self) -> NXGraphT:
        """Get an immutable view of the NetworkX graph without boundaries."""
        subgraph_view = nx.subgraph_view(
            self.graph, filter_node=lambda node: node not in self.boundaries
        )
        return cast(NXGraphT, subgraph_view)

    @cached_property
    def nodes(self) -> list[int]:
        return sorted(self.graph.nodes)

    @property
    def boundaries(self) -> frozenset[int]:
        """Return all nodes of this graph that are labelled as boundaries."""
        return self._boundaries

    @cached_property
    def detector_records(self) -> dict[int, DetectorRecord]:
        return dict(self.graph.nodes.data(data=True))

    @property
    def adj(self) -> nx.coreviews.MultiAdjacencyView:
        """
        Graph adjacency object holding the neighbours of each node.
        The returned object is a mappable that holds for each node
        a dictionary
        containing all of its neighbour nodes and the associated edge
        metadata (EdgeRecords).
        """
        return self._graph.adj

    def neighbors(self, detector: int) -> Iterator[int]:
        yield from self._graph[detector]

    def detector_is_boundary(self, detector: int) -> bool:
        """Return True if given detector is a boundary, False otherwise."""
        return self._boundaries_lookup[detector]

    def get_relevant_nodes(self, logicals: Iterable[set[DecodingEdge]]) -> set[int]:
        """Return the nodes that have an edge with a path to logical that
        is not via the boundary"""
        components = list(nx.connected_components(self.no_boundary_view))
        relevant_nodes: set[int] = set()
        for logical in logicals:
            relevant_nodes.update(chain.from_iterable(logical))
            for component in components:
                if any(a in component or b in component for a, b in logical):
                    relevant_nodes.update(component)
        return relevant_nodes

    def shortest_path(self, origin: int, destination: int) -> list[DecodingEdge]:
        """Find the shortest path between two syndrome bits, as a sequence of decoding
        edges.
        """
        # pylint: disable=too-many-function-args
        # in Python 3.10 use pairwise here
        firsts, seconds = tee(nx.shortest_path(self._graph, origin, destination))
        next(seconds)
        return [DecodingEdge(u, v) for u, v in zip(firsts, seconds, strict=False)]

    def shortest_path_length(self, origin: int, destination: int) -> float:
        """Find the length of the shortest path between two syndrome bits."""
        return nx.shortest_path_length(self._graph, origin, destination)

    def shortest_path_no_boundaries(
        self, origin: int, destination: int
    ) -> list[DecodingEdge]:
        """Find the shortest path between two syndrome bits without going via any
        boundaries, as a sequence of decoding edges.
        If origin or destination are boundaries, an exception will be raised.
        """
        # pylint: disable=too-many-function-args
        # in Python 3.10 use pairwise here
        firsts, seconds = tee(
            nx.shortest_path(self.no_boundary_view, origin, destination)
        )
        next(seconds)
        return [DecodingEdge(*edge) for edge in zip(firsts, seconds, strict=False)]

    def shortest_path_length_no_boundaries(
        self, origin: int, destination: int
    ) -> float:
        """Find the length of the shortest path between two syndrome bits without going
        via any boundaries.
        If origin or destination are boundaries, an exception will be raised.
        """
        return nx.shortest_path_length(self.no_boundary_view, origin, destination)


NXLogicals = Sequence[AbstractSet[DecodingEdge]]


@dataclass
class NXCode:
    """Convenience class encapsulating
    the graph and logicals."""

    graph: NXDecodingGraph
    logicals: NXLogicals


class NXDecodingMultiGraph(NXGraph[_QECNXMG, tuple[DecodingEdge, int]]):
    """Implementation of a decoding multigraph using NetworkX and QEC syndrome
    and edge objects. All edges in this graph are size 2, and a pair of syndromes
    can be connected by multiple edges (a multi-edge). As with the non-multiedge
    decoding graph, a fictitious boundary vertex exists to match isolated vertices.

    Parameters
    ----------
    graph : _QECNXMG
        NetworkX MultiGraph to wrap.
    boundaries : Optional[Iterable[int]], optional
        Optional indices of detectors that should be treated as a boundary, by default
        frozenset().
    """

    base_graph_class: ClassVar = _QECNXMG

    def __init__(self, graph: _QECNXMG, boundaries: Iterable[int] = frozenset()):
        super().__init__(graph, boundaries)

    @classmethod
    def from_edge_list(
        cls,
        edge_data: Iterable[
            DecodingEdge | tuple[int, int] | tuple[DecodingEdge, EdgeRecord]
        ],
        detector_records: dict[int, DetectorRecord] | None = None,
        boundaries: Iterable[int] = frozenset(),
    ) -> NXDecodingMultiGraph:
        """Create a graph, where connectivity is defined by an edge list.
        Other information about the graph is given via an optional list of nodes,
        boundaries and detector records.

        Parameters
        ----------
        edge_data : Iterable[Union[DecodingEdge, Tuple[int, int], Tuple[DecodingEdge, EdgeRecord]]],
            Collection of decoding edges to include in the graph, as an iterable of
            edges or as a Tuple of edges and their associated EdgeRecords.

        detector_records : Optional[Dict[int, DetectorRecord]], optional
            Optional specification of graph vertices given as a dictionary
            of detector index to detector record.

        edge_records : Optional[Dict[Tuple[DecodingEdge, int] | DecodingEdge, EdgeRecord]] = None, optional
            Optional specification of graph edges given as a dictionary
            of DecodingEdge to edge record.

        boundaries : Optional[Iterable[int]], optional
            Collection of nodes that are classified as a boundary vertex.

        Returns
        -------
        NXDecodingMultiGraph
            Constructed decoding multigraph.
        """  # noqa: E501
        nx_graph = cls.base_graph_class()
        detector_records = {} if detector_records is None else detector_records
        edge_records: dict[tuple[DecodingEdge, int], EdgeRecord] = {}
        nx_graph.add_nodes_from(detector_records.items())

        for data in edge_data:
            if isinstance(data, DecodingEdge):
                edge = data
                edge_record = EdgeRecord()
            elif _is_tuple_of_ints(data):
                edge = DecodingEdge(*data)
                edge_record = EdgeRecord()
            elif _is_decoding_edge_and_edge_record(data):
                edge, edge_record = data
            else:
                msg = f"Unsupported data type for edge {data}"
                raise TypeError(msg)
            u, v = edge
            k = nx_graph.number_of_edges(u, v)
            edge_records[(edge, k)] = edge_record
            nx_graph.add_edge(u, v, k, **edge_record)

        return cls(nx_graph, boundaries)

    @cached_property
    def edges(self) -> list[tuple[DecodingEdge, int]]:
        return [(DecodingEdge(u, v), edge_id) for u, v, edge_id in self.graph.edges]

    @cached_property
    def boundary_edges(self) -> set[tuple[DecodingEdge, int]]:
        return set(
            chain.from_iterable(
                self.incident_edges(boundary) for boundary in self.boundaries
            )
        )

    @cached_property
    def edge_records(self) -> dict[tuple[DecodingEdge, int], EdgeRecord]:
        return {
            (DecodingEdge(u, v), k): record
            for (u, v, k), record in self._graph.edges.items()
        }

    def get_edges(self, *detectors: int) -> Iterator[tuple[DecodingEdge, int]]:
        edge_data = self.graph.get_edge_data(*detectors)
        for edge_id in edge_data:  # type: ignore[union-attr]
            yield (DecodingEdge(*detectors), int(edge_id))

    def incident_edges(self, detector: int) -> Iterator[tuple[DecodingEdge, int]]:
        unique_detectors = {tuple(edge) for edge in self.graph.edges(detector)}
        for detectors in unique_detectors:
            yield from self.get_edges(*detectors)

    def to_parity_check_matrix(self) -> npt.NDArray[np.uint8]:
        check_matrix = np.zeros((len(self.nodes), len(self.edges)), dtype=np.uint8)
        for edge_index, (edge, _) in enumerate(self.edges):
            check_matrix[tuple(edge), (edge_index,)] = True
        return check_matrix

    def with_multi_edges_merged(
        self,
        combine_edge_records: Callable[
            [Iterable[EdgeRecord]], EdgeRecord
        ] = convolve_probabilities,
    ) -> NXDecodingGraph:
        """Create an `NXDecodingGraph` instance with the same nodes but with
        only single edges, merging the multi-edge edge records to a single edge
        record.

        Parameters
        ----------
        combine_edge_records : Callable[[ Iterable[EdgeRecord]], EdgeRecord], optional
            A function which combines all the edge records for a given edge,
            by default convolve_probabilities which applies the generalised
            `p * (1 - q) + q * (1 - p)`.

        Returns
        -------
        NXDecodingGraph
            A graph with the same properties as this graph but with all
            multi-edges combined to a single edge.
        """
        grouped_edge_records = defaultdict(list)
        for (edge, _), edge_record in self.edge_records.items():
            grouped_edge_records[edge].append(edge_record)
        new_graph = NXDecodingGraph(_QECNX(self._graph), self.boundaries)
        for edge, edge_records in grouped_edge_records.items():
            new_graph.edge_records[edge] = combine_edge_records(edge_records)
        return new_graph

    def error_to_syndrome(
        self, edges: Iterable[tuple[DecodingEdge, int]]
    ) -> OrderedSyndrome:
        return OrderedSyndrome(
            symptom
            for symptom in chain.from_iterable(edge for edge, _ in edges)
            if symptom not in self.boundaries
        )


class NXDecodingGraph(NXGraph[_QECNX, DecodingEdge]):
    """Implementation of a decoding graph using NetworkX and the QEC syndrome
    and edge objects. This works much like the hypergraph, but all edges are size 2.
    To account for boundaries, which in a hypergraph exist as size 1 edges, extra nodes
    may be added. These are virtual detectors that do not physically exist when
    running the syndrome extraction circuit.

    Parameters
    ----------
    graph : _QECNX
        NetworkX graph to wrap.
    boundaries : Optional[Iterable[int]], optional
        Optional indices of detectors that should be treated as a boundary, by default
        None.
    """

    base_graph_class: ClassVar = _QECNX

    def __init__(self, graph: _QECNX, boundaries: Iterable[int] = frozenset()):
        super().__init__(graph, boundaries)

    @classmethod
    def from_edge_list(
        cls,
        edge_data: Iterable[
            DecodingEdge | tuple[int, int] | tuple[DecodingEdge, EdgeRecord]
        ],
        detector_records: dict[int, DetectorRecord] | None = None,
        boundaries: Iterable[int] = frozenset(),
    ) -> NXDecodingGraph:
        """Create a graph, where connectivity is defined by an edge list.
        Other information about the graph is given via an optional list of nodes,
        boundaries and detector records.

        Parameters
        ----------
        edge_data : Iterable[Union[DecodingEdge, Tuple[int, int], Tuple[DecodingEdge, EdgeRecord]]],
            Collection of decoding edges to include in the graph, as an iterable of
            edges or as a Tuple of edges and their associated EdgeRecords.

        detector_records : Optional[Dict[int, DetectorRecord]], optional
            Optional specification of graph vertices given as a dictionary
            of detector index to detector record.

        boundaries : Optional[Iterable[int]], optional
            Collection of nodes that are classified as a boundary vertex.

        Returns
        -------
        NXDecodingGraph
            Constructed decoding graph.
        """  # noqa: E501
        nx_graph = cls.base_graph_class()
        detector_records = {} if detector_records is None else detector_records
        edge_records: dict[DecodingEdge, EdgeRecord] = {}
        nx_graph.add_nodes_from(detector_records.items())

        for data in edge_data:
            if isinstance(data, DecodingEdge):
                edge = data
                edge_records[edge] = EdgeRecord()
            elif isinstance(data, tuple) and isinstance(data[1], int):
                edge = DecodingEdge(*data)
                edge_records[edge] = EdgeRecord()
            elif isinstance(data, tuple) and isinstance(data[1], EdgeRecord):
                edge, edge_record = data
                edge_records[edge] = edge_record
            else:
                msg = f"Unsupported data type for edge {data}"
                raise TypeError(msg)
            nx_graph.add_edge(*edge, **edge_records[edge])

        return cls(nx_graph, boundaries)

    def to_decoding_hypergraph(self) -> DecodingHyperGraph:
        """Elevate decoding graph to hypergraph."""
        edge_data = [
            (DecodingHyperEdge(set(edge) - self.boundaries), self.edge_records[edge])
            for edge in self.edges
        ]
        det_records_copy = {
            k: v for k, v in self.detector_records.items() if k not in self.boundaries
        }
        return DecodingHyperGraph(edge_data, detector_records=det_records_copy)

    def get_edges(self, *detectors: int) -> list[DecodingEdge]:
        return [edge] if (edge := DecodingEdge(*detectors)) in self.edges else []

    def get_edge(self, *detectors: int) -> DecodingEdge:
        """Returns the unique edge in the graph corresponding to the given set
        of detector indices.

        Raises
        ------
        ValueError
            If no edge in the graph corresponds to the given detectors.
        """
        if (edge := DecodingEdge(*detectors)) in self.edge_records:
            return edge
        msg = f"Detectors {detectors} do not belong to any edge in the graph."
        raise ValueError(msg)

    def get_edge_record(self, *detectors: int) -> EdgeRecord:
        """Given a set of detectors that define an edge in the graph,
        return its correspoiding `EdgeRecord`.
        """
        return self.edge_records[DecodingEdge(*detectors)]

    @cached_property
    def edges(self) -> list[DecodingEdge]:
        return [DecodingEdge(u, v) for u, v in self._graph.edges]

    @cached_property
    def boundary_edges(self) -> set[DecodingEdge]:
        return set(
            chain.from_iterable(
                self.incident_edges(boundary) for boundary in self.boundaries
            )
        )

    @cached_property
    def edge_records(self) -> dict[DecodingEdge, EdgeRecord]:
        return {
            DecodingEdge(u, v): record for (u, v), record in self._graph.edges.items()
        }

    def incident_edges(self, detector: int) -> Iterator[DecodingEdge]:
        for edge in self._graph.edges(detector):
            yield DecodingEdge(*edge)

    def to_parity_check_matrix(self) -> npt.NDArray[np.uint8]:
        check_matrix = np.zeros((len(self.nodes), len(self.edges)), dtype=np.uint8)
        for edge_index, edge in enumerate(self.edges):
            check_matrix[tuple(edge), (edge_index,)] = True
        return check_matrix

    def error_to_syndrome(self, edges: Iterable[DecodingEdge]) -> OrderedSyndrome:
        return OrderedSyndrome(
            symptom
            for symptom in chain.from_iterable(edges)
            if symptom not in self.boundaries
        )



# (c) Copyright Riverlane 2020-2025.
from __future__ import annotations

from collections.abc import Iterator
from pathlib import Path

import numpy as np

from deltakit_core.decoding_graphs import Bitstring, OrderedSyndrome


def to_bytearray(data: Path | bytes) -> bytearray:
    """Converts either the content of a file or a bytes object to
    a bytearray object. A key motivation for this is that np.unpackbits
    accepts bytearray but not bytes."""
    if isinstance(data, Path):
        with data.open("rb") as b8_data:
            return bytearray(b8_data.read())
    return bytearray(data)


def b8_to_syndromes(
    b8_input: Path | bytes, detector_num: int
) -> Iterator[OrderedSyndrome]:
    """Given a b8 input (either a file path containing b8 data or bytes),
    and the number of detectors that should be in each syndrome, return a
    generator of syndromes.

    Informed by https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md#b8
    """
    bytes_per_shot = (detector_num + 7) // 8
    b8_byte_arr: bytearray = to_bytearray(b8_input)
    for offset in range(0, len(b8_byte_arr), bytes_per_shot):
        yield OrderedSyndrome.from_bitstring(
            np.unpackbits(
                np.frombuffer(
                    b8_byte_arr[offset : offset + bytes_per_shot], dtype=np.uint8
                ),
                bitorder="little",
            ).tolist()
        )


def b8_to_measurements(b8_input: Path, measurement_num: int) -> Iterator[Bitstring]:
    """Given a b8 input (a file path containing b8 data),
    and the number of measurements in each shot,
    return a generator of measurement bitstrings.

    Informed by https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md#b8
    """
    bytes_per_shot = (measurement_num + 7) // 8
    b8_byte_arr: bytearray = to_bytearray(b8_input)
    for offset in range(0, len(b8_byte_arr), bytes_per_shot):
        yield Bitstring.from_bits(
            np.unpackbits(
                np.frombuffer(
                    b8_byte_arr[offset : offset + bytes_per_shot], dtype=np.uint8
                ),
                bitorder="little",
            ).tolist()
        )


def b8_to_logical_flip(
    b8_input: Path | bytes, num_logicals: int = 1
) -> Iterator[tuple[bool, ...]]:
    """Given a b8 input (either a file path containing b8 data or bytes),
    and the number of detectors that should be in each syndrome, return a
    generator of logical flips.

    Informed by https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md#b8
    """
    bytes_per_shot = (num_logicals + 7) // 8
    b8_byte_arr: bytearray = to_bytearray(b8_input)
    for offset in range(0, len(b8_byte_arr), bytes_per_shot):
        yield tuple(
            np.unpackbits(
                np.frombuffer(
                    b8_byte_arr[offset : offset + bytes_per_shot], dtype=np.uint8
                ),
                bitorder="little",
            )[:num_logicals].tolist(),
        )


def syndromes_to_b8_file(
    syndrome_b8_out: Path, detector_num: int, syndromes: Iterator[OrderedSyndrome]
):
    """Given a syndrome generator, this will output them to file in the b8 format.
    Informed by https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md#b8
    """
    with syndrome_b8_out.open("wb") as b8_out_handle:
        b8_out_handle.writelines(
            bytes(np.packbits(syndrome.as_bitstring(detector_num), bitorder="little"))
            for syndrome in syndromes
        )


def logical_flips_to_b8_file(
    logical_flips_b8_out: Path, logical_flips: Iterator[tuple[bool, ...]]
):
    """Given a logical flip generator, this will output them to file in the b8 format.
    b8 is a byte-padded bitstring representation for logical flips. This is more
    efficient than the string-based 01 representation.
    Informed by https://github.com/quantumlib/Stim/blob/main/doc/result_formats.md#b8
    """
    with logical_flips_b8_out.open("wb") as data_b8:
        data_b8.writelines(
            bytes(np.packbits(logical_flip, bitorder="little"))
            for logical_flip in logical_flips
        )


def parse_01_to_logical_flips(
    logical_flips_01_file: Path,
) -> Iterator[tuple[bool, ...]]:
    """Given a filepath for a 01 file containing target logical flips, return a
    generator of logical flips as booleans.

    01 files are string based, where each line is a different shot. Where 1
    indicates the reference logical is flipped.

    Use logical_flips_to_b8_file on the result of this function to translate 01
    data to b8 data that can be decoded via the B8DecoderManager.
    """
    with logical_flips_01_file.open("r", encoding="utf-8") as logical_flips_handle:
        for bit_string in logical_flips_handle.read().splitlines():
            yield tuple(char != "0" for char in bit_string)


def parse_01_to_syndromes(
    logical_flips_01_file: Path,
) -> Iterator[OrderedSyndrome]:
    """Given a filepath for a 01 file containing syndromes, return a
    generator of OrderedSyndromes.

    01 files are string based, where each line is a different target logical as a
    string of 0s and 1s. Where 1 indicates the reference logical is flipped.

    Use logical_flips_to_b8_file on the result of this function to translate 01
    data to b8 data that can be decoded via the B8DecoderManager.
    """
    with logical_flips_01_file.open("r", encoding="utf-8") as logical_flips_handle:
        for bit_string in logical_flips_handle.read().splitlines():
            yield OrderedSyndrome([i for i, v in enumerate(bit_string) if v == "1"])



from enum import Enum
from typing import Final


class RiverlaneColors(Enum):
    """Riverlane colors for visualisation."""

    LIGHT_GREEN = "#00968f"
    RUST = "#dc4405"
    DARK_GREEN = "#1d3c34"
    BLUE_70_PERC = "#77dbe5"
    ORANGE = "#ff7500"
    MILD_GREEN = "#006f62"
    BLUE = "#3ccbda"
    ORANGE_70_PERC = "#ff9e4d"
    BLACK_70_PERC = "#777776"


RIVERLANE_PLOT_COLOURS: Final[list[str]] = [
    "#006F62",
    "#FF6A00",
    "#4B5BFF",
    "#E01500",
    "#003B40",
    "#D835A2",
    "#00A3A3",
    "#9B3EC6",
    "#54170E",
    "#FFD00B",
    "#0AD2F2",
]
RIVERLANE_DARK_TEXT_COLOUR = "#0A1600"
RIVERLANE_WHITE_TEXT_COLOUR = "#E5F0EF"
RIVERLANE_LINE_COLOUR = "#003B40"
RIVERLANE_NEUTRAL_GREY = "#DBDBDB"



# (c) Copyright Riverlane 2020-2025.
"""Module for functions for weighted decoding graphs."""

from collections import defaultdict
from collections.abc import Sequence

import numpy as np

from deltakit_core.decoding_graphs._decoding_graph import (
    DecodingEdge,
    EdgeRecord,
    NXDecodingGraph,
)


def change_graph_error_probabilities(
    graph: NXDecodingGraph, new_p_errors: Sequence[float]
) -> NXDecodingGraph:
    """Make a new graph from the given graph where the error probabilities on the edges
    are changed to the new error probabilities. The sequence of new error probabilities
    should correspond to the sequence of edges in the given graph.

    Parameters
    ----------
    graph : NXDecodingGraph
        The graph to use to with the old error probabilities.
    new_p_errors : Sequence[float]
        The new error probabilities to apply to the graph.

    Returns
    -------
    NXDecodingGraph
        A new graph instance with the same vertices, edges and boundaries as
        the input graph but with an updated error proability on each edge.

    Raises
    ------
    ValueError
        If the length of the new error probabilities and edges do not match.
    """
    if len(graph.edges) != len(new_p_errors):
        msg = (
            "There should be an equal number of new error probabilities to edges in "
            "the graph."
        )
        raise ValueError(msg)
    return NXDecodingGraph.from_edge_list(
        [
            (DecodingEdge(u, v), EdgeRecord(p_err=p_err))
            for (u, v), p_err in zip(graph.edges, new_p_errors, strict=False)
        ],
        detector_records=graph.detector_records,
        boundaries=graph.boundaries,
    )


def vector_weights(
    graph: NXDecodingGraph,
) -> defaultdict[tuple[float, ...], list[float]]:
    """Treat the edges of a graph as vectors and return a mapping of
    normalised vectors to a list of weights on those vectors.

    Parameters
    ----------
    graph : NXDecodingGraph
        The edges on this graph represent vectors.

    Returns
    -------
    DefaultDict[Tuple[float, ...], List[float]]
    """
    vectors = defaultdict(list)
    for edge in graph.edges:
        u, v = edge
        edge_record = graph.edge_records[edge]
        if (u_coord := graph.detector_records[u].full_coord) == (0,):
            msg = f"{u} does not have proper coordinates."
            raise ValueError(msg)
        if (v_coord := graph.detector_records[v].full_coord) == (0,):
            msg = f"{v} does not have proper coordinates."
            raise ValueError(msg)
        if u in graph.boundaries or v in graph.boundaries:
            vectors[(0.0,) * len(u_coord)].append(edge_record.weight)
        else:
            vector = np.subtract(v_coord, u_coord)
            # Take the convention that the first non-zero element of the
            # vector must be positive.
            vector *= np.sign(vector[vector != 0][0])
            vectors[tuple(vector.tolist())].append(edge_record.weight)
    return vectors



# (c) Copyright Riverlane 2020-2025.
"""
Sub-package for data formats and converting them to other data types
"""

from deltakit_core.data_formats._b801_parsers import (
    b8_to_logical_flip,
    b8_to_measurements,
    b8_to_syndromes,
    logical_flips_to_b8_file,
    parse_01_to_logical_flips,
    parse_01_to_syndromes,
    syndromes_to_b8_file,
    to_bytearray,
)
from deltakit_core.data_formats._measurements import (
    c64_to_addressed_input_words,
    split_input_data_to_c64,
)

# List only public members in `__all__`.
__all__ = [
    "b8_to_logical_flip",
    "b8_to_measurements",
    "b8_to_syndromes",
    "c64_to_addressed_input_words",
    "logical_flips_to_b8_file",
    "parse_01_to_logical_flips",
    "parse_01_to_syndromes",
    "split_input_data_to_c64",
    "syndromes_to_b8_file",
    "to_bytearray",
]



import functools
import warnings
from collections.abc import Callable
from typing import ParamSpec, TypeVar, overload

from packaging.version import Version

P = ParamSpec("P")
RetType = TypeVar("RetType")


@overload
def deprecated(f: Callable[P, RetType] | None, /) -> Callable[P, RetType]:
    pass  # pragma: no cover


@overload
def deprecated(
    *,
    reason: str | None = None,
    replaced_by: str | None = None,
    removed_in_version: Version | None = None,
) -> Callable[[Callable[P, RetType]], Callable[P, RetType]]:
    pass  # pragma: no cover


def deprecated(
    f: Callable[P, RetType] | None = None,
    /,
    *,
    reason: str | None = None,
    replaced_by: str | None = None,
    removed_in_version: Version | None = None,
) -> Callable[[Callable[P, RetType]], Callable[P, RetType]] | Callable[P, RetType]:
    """Deprecation decorator.

    Args:
        f: function to deprecate, allow this function to be used as a decorator.
        reason: rationale behind the deprecation.
        replaced_by: a description of the new method that should be used instead of the
            deprecated function.
        removed_in_version: minimum version in which the deprecated function might be
            removed. The deprecated function should not be removed earlier than this
            version.

    Returns:
        Depending on the provided arguments, either a wrapper around the provided ``f``
        that raises a deprecation warning or a decorator that can deprecate a function.
    """
    if all(p is None for p in (f, reason, replaced_by, removed_in_version)):
        msg = (
            "Expected at least one of f, reason, replaced_by or removed_in_version to "
            "be provided. All of them are None."
        )
        raise ValueError(msg)
    if f is not None:
        msg = f"{f.__name__} is deprecated and will eventually be removed."

        @functools.wraps(f)
        def deprecated_func(*args: P.args, **kwargs: P.kwargs) -> RetType:
            warnings.warn(msg, DeprecationWarning, stacklevel=1)
            return f(*args, **kwargs)

        return deprecated_func

    # else
    def deprecation_decorator(
        func_to_deprecate: Callable[P, RetType],
    ) -> Callable[P, RetType]:
        msg = f"{func_to_deprecate.__name__} is deprecated and will "
        msg += (
            f"be removed in version {removed_in_version}."
            if removed_in_version is not None
            else "eventually be removed."
        )
        if reason is not None:
            msg += f" Reason for deprecation: '{reason}'."
        if replaced_by is not None:
            msg += f" Consider using '{replaced_by}' instead."

        @functools.wraps(func_to_deprecate)
        def deprecated_func(*args: P.args, **kwargs: P.kwargs) -> RetType:
            warnings.warn(msg, DeprecationWarning, stacklevel=1)
            return func_to_deprecate(*args, **kwargs)

        return deprecated_func

    return deprecation_decorator



# (c) Copyright Riverlane 2020-2025.
"""Utility functions for windowing."""

from __future__ import annotations

from collections import deque
from collections.abc import Iterable

from deltakit_core.decoding_graphs._data_qubits import EdgeRecord
from deltakit_core.decoding_graphs._decoding_graph import (
    DecodingHyperEdge,
    DecodingHyperGraph,
)


def nodes_within_radius(
    hypergraph: DecodingHyperGraph, start_nodes: Iterable[int], radius: int
) -> set[int]:
    """Compute nodes within a given unweighted graph distance from any start node.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        The global decoding hypergraph.
    start_nodes : Iterable[int]
        Nodes from which to grow the region.
    radius : int
        Number of unweighted adjacency steps to grow.

    Returns
    -------
    Set[int]
        Set of detector node ids within the specified hop distance of any start node.

    """
    if radius < 0:
        msg = "Radius must be non-negative"
        raise ValueError(msg)
    visited: set[int] = set(start_nodes)
    if radius == 0:
        return visited
    frontier: deque[tuple[int, int]] = deque((n, 0) for n in visited)
    while frontier:
        node, dist = frontier.popleft()
        if dist == radius:
            continue
        for neighbour in hypergraph.neighbors(node):
            if neighbour not in visited:
                visited.add(neighbour)
                frontier.append((neighbour, dist + 1))
    return visited


def weighted_nodes_within_radius(
    hypergraph: DecodingHyperGraph, start_nodes: Iterable[int], radius: float
) -> set[int]:
    """Compute nodes within a weighted distance budget from any start node.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        The global decoding hypergraph.
    start_nodes : Iterable[int]
        Seed nodes to begin weighted growth.
    radius : float
        Maximum accumulated edge weight distance.

    Returns
    -------
    Set[int]
        Nodes reachable with total path weight <= ``radius``.
    """
    msg = "Weighted growth is not yet implemented."
    raise NotImplementedError(msg)


def induce_subhypergraph(
    hypergraph: DecodingHyperGraph, nodes: set[int]
) -> DecodingHyperGraph:
    """Form the induced sub-hypergraph on the provided set of nodes. Keeps only
    hyperedges whose vertices are fully contained in ``nodes`` and restricts detector
    records accordingly.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Source decoding hypergraph.
    nodes : Set[int]
        Set of hyergraph nodes to induce subgraph from.

    Returns
    -------
    DecodingHyperGraph
        New hypergraph containing only edges fully within ``nodes`` and detector records
        for those nodes.
    """
    edge_data: list[tuple[DecodingHyperEdge, EdgeRecord]] = []
    for edge, record in hypergraph.edge_records.items():
        if edge.vertices.issubset(nodes):
            edge_data.append((edge, record))
    det_records = {n: hypergraph.detector_records[n] for n in nodes}
    return DecodingHyperGraph(edge_data, detector_records=det_records)


def connect_dangling_to_boundary_hypergraph(
    supergraph: DecodingHyperGraph, subgraph: DecodingHyperGraph
) -> DecodingHyperGraph:
    """Fold edges leaving the subgraph into unary boundary edges per node.

    For each node in ``subgraph``, consider all edges in ``supergraph`` that include the
    node and at least one vertex outside ``subgraph``. These excluded edges are folded
    into a single unary hyperedge on that node with ``p_err`` equal to the combined
    probability of their probabilities. If a unary edge already exists on the node in
    the subgraph, its probability is combined with the folded probability.

    Parameters
    ----------
    supergraph : DecodingHyperGraph
        The original global hypergraph.
    subgraph : DecodingHyperGraph
        The induced subgraph to which boundary edges will be added.

    Returns
    -------
    DecodingHyperGraph
        A new hypergraph with updated/added unary boundary edges and the same detector
        records as ``subgraph``.
    """

    def _combine_probabilities(prob1: float, prob2: float) -> float:
        """Helper to combine two edge probabilities."""
        return prob1 * (1 - prob2) + (1 - prob1) * prob2

    nodes = set(subgraph.nodes)
    folded_records = dict(subgraph.edge_records)

    for node in nodes:
        folded_p = 0.0
        for edge in supergraph.incident_edges(node):
            if not edge.vertices.issubset(nodes):
                folded_p = _combine_probabilities(
                    folded_p, supergraph.edge_records[edge].p_err
                )
        if folded_p == 0.0:
            continue
        unary_edge = DecodingHyperEdge((node,))
        if unary_edge in folded_records:
            existing_p = folded_records[unary_edge].p_err
            folded_records[unary_edge] = EdgeRecord(
                p_err=_combine_probabilities(existing_p, folded_p)
            )
        else:
            folded_records[unary_edge] = EdgeRecord(p_err=folded_p)

    edge_data = list(folded_records.items())
    return DecodingHyperGraph(edge_data, detector_records=subgraph.detector_records)


def relabel_hypergraph_nodes_contiguously(
    hypergraph: DecodingHyperGraph,
) -> tuple[DecodingHyperGraph, dict[int, int]]:
    """Return a new hypergraph with nodes relabelled to a contiguous 0..N-1 range.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        The hypergraph whose node labels may be sparse
        relative to global indices.

    Returns
    -------
    Tuple[DecodingHyperGraph, Dict[int, int]]
        A tuple of (relabelled_hypergraph, global_to_window_detector_map) where
        the mapping gives global node id -> window id.
    """
    node_mapping: dict[int, int] = {
        old: i for i, old in enumerate(sorted(hypergraph.nodes))
    }

    remapped_edge_data: list[tuple[DecodingHyperEdge, EdgeRecord]] = []
    for edge, record in hypergraph.edge_records.items():
        remapped_edge = DecodingHyperEdge(tuple(node_mapping[n] for n in edge))
        remapped_edge_data.append((remapped_edge, record))

    remapped_detector_records = {
        node_mapping[n]: rec
        for n, rec in hypergraph.detector_records.items()
        if n in node_mapping
    }

    relabelled = DecodingHyperGraph(
        remapped_edge_data, detector_records=remapped_detector_records
    )
    return relabelled, node_mapping


def expand_nodes_to_time_span(
    hypergraph: DecodingHyperGraph, nodes: set[int]
) -> set[int]:
    """Expand ``nodes`` so that for every time value represented among ``nodes`` all
    detectors in the hypergraph with that time are included.

    Parameters
    ----------
    hypergraph : DecodingHyperGraph
        Hypergraph providing detector records.
    nodes : Set[int]
        Initial grown node set in hypergraph.

    Returns
    -------
    Set[int]
        Expanded node set containing full rounds.
    """
    if not nodes:
        return set()
    times = [hypergraph.detector_records[n].time for n in nodes]
    t_min, t_max = min(times), max(times)
    return {
        n
        for n, rec in hypergraph.detector_records.items()
        if t_min <= rec.time <= t_max
    }



# (c) Copyright Riverlane 2020-2025.
"""
Sub-package for defining decoding graphs and related data types; these structures are
used by many decoders to define their understanding of the code and noise model being
decoded.
"""

from deltakit_core.decoding_graphs._data_qubits import (
    DecodingEdge,
    DecodingHyperEdge,
    EdgeRecord,
    EdgeT,
    OrderedDecodingEdges,
    errors_to_syndrome,
)
from deltakit_core.decoding_graphs._decoding_graph import (
    AnyEdgeT,
    DecodingCode,
    DecodingGraphT,
    DecodingHyperGraph,
    DecodingHyperMultiGraph,
    HyperLogicals,
    HyperMultiGraph,
    NXCode,
    NXDecodingGraph,
    NXDecodingMultiGraph,
    NXGraph,
    NXGraphT,
    NXLogicals,
    convolve_probabilities,
)
from deltakit_core.decoding_graphs._decoding_graph_tools import (
    compute_graph_distance,
    compute_graph_distance_for_logical,
    filter_to_data_edges,
    filter_to_measure_edges,
    graph_to_json,
    has_contiguous_nodes,
    hypergraph_to_weighted_edge_list,
    inverse_logical_at_boundary,
    is_logical_along_boundary,
    is_single_connected_component,
    nx_graph_from_json,
    parse_stim_circuit,
    single_boundary_is_last_node,
    stim_circuit_to_graph_dem,
    unweight_graph,
    worst_case_num_detectors,
)
from deltakit_core.decoding_graphs._dem_parsing import (
    CoordinateOffset,
    DemParser,
    DetectorCounter,
    DetectorHandler,
    DetectorRecorder,
    ErrorHandler,
    LogicalsInEdges,
    ObservableHandler,
    collect_edges,
    dem_to_decoding_graph_and_logicals,
    dem_to_hypergraph_and_logicals,
    observable_warning,
)
from deltakit_core.decoding_graphs._explained_dem_parsing import (
    depolarising_as_independent,
    extract_logicals,
    noise_probability,
    parse_explained_dem,
)
from deltakit_core.decoding_graphs._hypergraph_annotations_tools import (
    annotate_edges_with_window_ids,
    get_unique_logical_window_ids,
    get_unique_window_ids,
    get_window_id_transfer_graph,
    separate_edges_per_window_id,
    separate_logicals_by_window_id,
)
from deltakit_core.decoding_graphs._hypergraph_decomposition import decompositions
from deltakit_core.decoding_graphs._syndromes import (
    Bitstring,
    DetectorRecord,
    FixedWidthBitstring,
    OrderedSyndrome,
    get_round_words,
)
from deltakit_core.decoding_graphs._weighted_graphs import (
    change_graph_error_probabilities,
    vector_weights,
)
from deltakit_core.decoding_graphs._windowing_utils import (
    connect_dangling_to_boundary_hypergraph,
    expand_nodes_to_time_span,
    induce_subhypergraph,
    nodes_within_radius,
    relabel_hypergraph_nodes_contiguously,
    weighted_nodes_within_radius,
)

# List only public members in `__all__`.
__all__ = [
    "AnyEdgeT",
    "Bitstring",
    "CoordinateOffset",
    "DecodingCode",
    "DecodingEdge",
    "DecodingGraphT",
    "DecodingHyperEdge",
    "DecodingHyperGraph",
    "DecodingHyperMultiGraph",
    "DemParser",
    "DetectorCounter",
    "DetectorHandler",
    "DetectorRecord",
    "DetectorRecorder",
    "EdgeRecord",
    "EdgeT",
    "ErrorHandler",
    "FixedWidthBitstring",
    "HyperLogicals",
    "HyperMultiGraph",
    "LogicalsInEdges",
    "NXCode",
    "NXDecodingGraph",
    "NXDecodingMultiGraph",
    "NXGraph",
    "NXGraphT",
    "NXLogicals",
    "ObservableHandler",
    "OrderedDecodingEdges",
    "OrderedSyndrome",
    "annotate_edges_with_window_ids",
    "change_graph_error_probabilities",
    "collect_edges",
    "compute_graph_distance",
    "compute_graph_distance_for_logical",
    "connect_dangling_to_boundary_hypergraph",
    "convolve_probabilities",
    "decompositions",
    "dem_to_decoding_graph_and_logicals",
    "dem_to_hypergraph_and_logicals",
    "depolarising_as_independent",
    "errors_to_syndrome",
    "expand_nodes_to_time_span",
    "extract_logicals",
    "filter_to_data_edges",
    "filter_to_measure_edges",
    "get_round_words",
    "get_unique_logical_window_ids",
    "get_unique_window_ids",
    "get_window_id_transfer_graph",
    "graph_to_json",
    "has_contiguous_nodes",
    "hypergraph_to_weighted_edge_list",
    "induce_subhypergraph",
    "inverse_logical_at_boundary",
    "is_logical_along_boundary",
    "is_single_connected_component",
    "nodes_within_radius",
    "noise_probability",
    "nx_graph_from_json",
    "observable_warning",
    "parse_explained_dem",
    "parse_stim_circuit",
    "relabel_hypergraph_nodes_contiguously",
    "separate_edges_per_window_id",
    "separate_logicals_by_window_id",
    "single_boundary_is_last_node",
    "stim_circuit_to_graph_dem",
    "unweight_graph",
    "vector_weights",
    "weighted_nodes_within_radius",
    "worst_case_num_detectors",
]



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.core`` namespace here."""

import importlib.metadata

# It looks like these were intended to be separate, public modules.
# For now, import them as such. This can be reconsidered during API review.
from . import decoding_graphs

__version__ = importlib.metadata.version(__package__)

# Prevent import of `importlib` (and any other non-public objects) from this module.
del importlib

# List only public members in `__all__`.
__all__ = ["decoding_graphs"]



# (c) Copyright Riverlane 2020-2025.
from collections import defaultdict

try:
    import lestim as stim
except ImportError:
    import stim

from deltakit_core.decoding_graphs._decoding_graph import (
    AnyEdgeT,
    DecodingHyperEdge,
    DecodingHyperMultiGraph,
    DetectorRecord,
    EdgeRecord,
    HyperMultiGraph,
)


def depolarising_as_independent(probability: float, num_qubits: int) -> float:
    """A depoloarising noise channel is equivalent to applying each of its
    component Pauli channels independently with a carefully chosen
    probability.  This function computes that probability.

    Parameters
    ----------
    probability : float
        The probability of the depolarising error occurring.
    num_qubits : int
        The number of qubits the depolarising channel acts on.

    Returns
    -------
    float
        The equivalent probability of applying Pauli errors independently.
    """
    # Need to convert a probability of applying X, Y or Z into a probability
    # of applying X, Y, Z or I.
    pauli_combinations = 4**num_qubits
    mixing_probability = (pauli_combinations - 1) / pauli_combinations
    if probability > mixing_probability:
        msg = (
            "Depolarising probability cannot be above the mixing "
            f"probability which is {mixing_probability}"
        )
        raise ValueError(msg)
    p_with_i = probability / mixing_probability
    exponent = 1 / 2 ** (2 * num_qubits - 1)
    return float((1 - (1 - p_with_i) ** exponent) / 2)


def noise_probability(noise_channel: stim.CircuitTargetsInsideInstruction) -> float:
    """Calculate the independent probability of a Pauli error happening from a
    given stim noise gate.

    Parameters
    ----------
    noise_channel : stim.CircuitTargetsInsideInstruction
        The stim noise channel

    Returns
    -------
    float
        The independent probability of a Pauli error.

    Raises
    ------
    TypeError
        If the gate type is not a noise channel or is a noise channel that
        cannot be expressed as independent error.
    """
    if noise_channel.gate == "DEPOLARIZE1":
        return depolarising_as_independent(noise_channel.args[0], 1)
    if noise_channel.gate == "DEPOLARIZE2":
        return depolarising_as_independent(noise_channel.args[0], 2)
    if noise_channel.gate in ("X_ERROR", "Y_ERROR", "Z_ERROR"):
        return float(noise_channel.args[0])
    msg = f"Unsupported gate type: {noise_channel.gate}"
    raise TypeError(msg)


def parse_explained_dem(
    explained_dem: list[stim.ExplainedError],
) -> DecodingHyperMultiGraph:
    """Parse an explained DEM into a hyper-multi-graph. Information about
    which logicals an edge affects is stored in the edge record of that edge

    Parameters
    ----------
    explained_dem : List[stim.ExplainedError]
        The explained DEM which comes from the
        `stim.Circuit.explain_detector_error_model_errors` method.

    Returns
    -------
    DecodingHyperMultiGraph
        A HyperMultiGraph object which represents all possible errors in the
        explained DEM.

    Raises
    ------
    TypeError
        If there is a target separator (^) in the explained DEM since we don't
        expect this character to appear there.
    """
    detector_records = {}
    edges = []
    for error in explained_dem:
        edge_vertices = set()
        logicals_affected = []
        for dem_term in error.dem_error_terms:
            target = dem_term.dem_target
            if target.is_separator():
                msg = "Target separators should not be in the explained DEM."
                raise TypeError(msg)
            if target.is_relative_detector_id():
                edge_vertices.add(target.val)
                *spatial_coords, time = dem_term.coords
                detector_records[target.val] = DetectorRecord(
                    tuple(spatial_coords), time=time
                )
            elif target.is_logical_observable_id():
                logicals_affected.append(target.val)

        edge = DecodingHyperEdge(edge_vertices)
        for error_location in error.circuit_error_locations:
            p_err = noise_probability(error_location.instruction_targets)
            edges.append(
                (edge, EdgeRecord(p_err=p_err, logicals_affected=logicals_affected))
            )

    return DecodingHyperMultiGraph(edges, detector_records=detector_records)


def extract_logicals(graph: HyperMultiGraph[AnyEdgeT]) -> list[set[AnyEdgeT]]:
    """Extract which edges affect which logical observables from a graph by
    looking through the edge records.

    Parameters
    ----------
    graph : HyperMultiGraph[AnyEdgeT]
        The graph containing edges records annotated with which edges affect
        which logicals.

    Returns
    -------
    List[Set[AnyEdgeT]]
        A list where each index represents a logical and the set of edges in
        at each index are the edges in the graph which affect that logical.
    """
    max_logical = 0
    logicals_dict = defaultdict(set)
    for edge, edge_record in graph.edge_records.items():
        for logical in edge_record.get("logicals_affected", []):
            max_logical = max(max_logical, logical)
            logicals_dict[logical].add(edge)

    # Logicals are zero indexed so we must plus one
    logicals: list[set[AnyEdgeT]] = [set() for _ in range(max_logical + 1)]
    for logical, edges in logicals_dict.items():
        logicals[logical] = edges

    return logicals



# (c) Copyright Riverlane 2020-2025.
from collections.abc import Iterator, Sequence

from deltakit_core.decoding_graphs import DecodingHyperEdge


def decompositions(
    target_edge: DecodingHyperEdge, edges: Sequence[DecodingHyperEdge]
) -> Iterator[frozenset[DecodingHyperEdge]]:
    """Generate decompositions of the given edge into the parts from edges.
    The union of all vertices in the parts must equal the vertices in the
    original edges for it to be considered a valid decomposition.

    Parameters
    ----------
    target_edge : DecodingHyperEdge
        The hyperedge to decompose into parts.
    edges : Sequence[DecodingHyperEdge]
        The edges in which to search for a decomposition.

    Yields
    ------
    Iterator[FrozenSet[DecodingHyperEdge]]
        A generator of frozensets with each item a different decomposition
        of the input edge.
    """

    # explore tree of possible decompositions

    def _decompositions(
        vertices: frozenset[int],
        legal_edges: list[DecodingHyperEdge],
        chosen_edges: frozenset[DecodingHyperEdge],
    ) -> Iterator[frozenset[DecodingHyperEdge]]:
        if len(vertices) == 0:  # done on this branch
            yield chosen_edges
            return
        if len(legal_edges) == 0:  # dead branch
            return
        first, *others = legal_edges
        if first.vertices.issubset(vertices):
            # use it
            yield from _decompositions(
                vertices - first.vertices, others, chosen_edges | {first}
            )
        # don't use it (even if you could)
        yield from _decompositions(vertices, others, chosen_edges)

    yield from _decompositions(
        target_edge.vertices,
        [edge for edge in edges if edge.vertices.issubset(target_edge)],
        frozenset(),
    )
